import numpy as np

import vel.util.math as math_util
import vel.util.tensor_util as tensor_util


class Rollout:
    """ Base class for environment rollout data """

    def to_transitions(self) -> 'Transitions':
        """ Convert given rollout to Transitions """
        raise NotImplementedError

    def episode_information(self):
        """ List of information about finished episodes """
        raise NotImplementedError

    def frames(self):
        """ Number of frames in rollout """
        raise NotImplementedError

    def batch_tensor(self, name):
        """ A buffer of a given value in a 'flat' (minibatch-indexed) format """
        raise NotImplementedError


class Transitions(Rollout):
    """
    Rollout of random transitions, that don't necessarily have to be in any order

    transition_tensors - tensors that have a row (multidimensional) per each transition. E.g. state, reward, done
    """
    def __init__(self, size, environment_information, transition_tensors, extra_data=None):
        self.size = size
        self.environment_information = environment_information
        self.transition_tensors = transition_tensors
        self.extra_data = extra_data

    def to_transitions(self) -> 'Transitions':
        """ Convert given rollout to Transitions """
        return self

    def episode_information(self):
        """ List of information about finished episodes """
        return [info.get('episode') for info in self.environment_information if 'episode' in info]

    def frames(self):
        """ Number of frames in this rollout """
        return self.size

    def shuffled_batches(self, batch_size):
        """ Generate randomized batches of data """
        if batch_size >= self.size:
            yield self
        else:
            batch_splits = math_util.divide_ceiling(self.size, batch_size)
            indices = list(range(self.size))
            np.random.shuffle(indices)

            for sub_indices in np.array_split(indices, batch_splits):
                yield Transitions(
                    len(sub_indices),
                    environment_information=None,
                    # Dont use it in batches for a moment, can be uncommented later if needed
                    # environment_information=[info[sub_indices.tolist()] for info in self.environment_information]
                    transition_tensors={k: v[sub_indices] for k, v in self.transition_tensors.items()}
                    # extra_data does not go into batches
                )

    def batch_tensor(self, name):
        """ A buffer of a given value in a 'flat' (minibatch-indexed) format """
        return self.transition_tensors[name]


class Trajectories(Rollout):
    """
    Rollout of trajectories - a number of consecutive transitions

    transition_tensors - tensors that have a row (multidimensional) per each transition. E.g. state, reward, done
    rollout_tensors - tensors that have a row (multidimensional) per whole rollout. E.g. final_value, initial rnn state
    """
    def __init__(self, num_steps, num_envs, environment_information, transition_tensors, rollout_tensors, extra_data=None):
        self.num_steps = num_steps
        self.num_envs = num_envs
        self.environment_information = environment_information
        self.transition_tensors = transition_tensors
        self.rollout_tensors = rollout_tensors
        self.extra_data = extra_data

    def to_transitions(self) -> 'Transitions':
        """ Convert given rollout to Transitions """
        # No need to propagate 'rollout_tensors' as they won't mean anything
        return Transitions(
            size=self.num_steps * self.num_envs,
            environment_information=[ei for l in self.environment_information for ei in l],
            transition_tensors={
                name: tensor_util.merge_first_two_dims(t) for name, t in self.transition_tensors.items()
            },
            extra_data=self.extra_data
        )

    def shuffled_batches(self, batch_size):
        """ Generate randomized batches of data - only sample whole trajectories """
        if batch_size >= self.num_envs * self.num_steps:
            yield self
        else:
            rollouts_in_batch = batch_size // self.num_steps

            batch_splits = math_util.divide_ceiling(self.num_envs, rollouts_in_batch)

            indices = list(range(self.num_envs))
            np.random.shuffle(indices)

            for sub_indices in np.array_split(indices, batch_splits):
                yield Trajectories(
                    num_steps=self.num_steps,
                    num_envs=len(sub_indices),
                    # Dont use it in batches for a moment, can be uncommented later if needed
                    # environment_information=[x[sub_indices.tolist()] for x in self.environment_information],
                    environment_information=None,
                    transition_tensors={k: x[:, sub_indices] for k, x in self.transition_tensors.items()},
                    rollout_tensors={k: x[sub_indices] for k, x in self.rollout_tensors.items()},
                    # extra_data does not go into batches
                )

    def batch_tensor(self, name):
        """ A buffer of a given value in a 'flat' (minibatch-indexed) format """
        if name in self.transition_tensors:
            return tensor_util.merge_first_two_dims(self.transition_tensors[name])
        else:
            return self.rollout_tensors[name]

    def flatten_tensor(self, tensor):
        """ Merge first two dims of a tensor """
        return tensor_util.merge_first_two_dims(tensor)

    def episode_information(self):
        """ List of information about finished episodes """
        return [
            info.get('episode') for infolist in self.environment_information for info in infolist if 'episode' in info
        ]

    def frames(self):
        """ Number of frames in rollout """
        return self.num_steps * self.num_envs



from torch.optim import Optimizer

from vel.api import TrainingInfo, EpochInfo, Learner
from vel.api.base import Model, Source


class TrainPhase:
    """ A single phase of training """

    @property
    def number_of_epochs(self) -> int:
        """ How many epochs does this phase take """
        raise NotImplementedError

    def set_up_phase(self, training_info: TrainingInfo, model: Model, source: Source) -> Optimizer:
        """ Prepare the phase for learning, returns phase optimizer """
        pass

    def restore(self, training_info: TrainingInfo, local_batch_idx: int, model: Model, hidden_state: dict):
        """
        Restore learning from intermediate state.
        """
        pass

    def epoch_info(self, training_info: TrainingInfo, global_idx: int, local_idx: int) -> EpochInfo:
        """ Create Epoch info """
        raise NotImplementedError

    def execute_epoch(self, epoch_info: EpochInfo, learner: Learner):
        """
        Execute epoch training.
        """
        raise NotImplementedError

    def tear_down_phase(self, training_info: TrainingInfo, model: Model):
        """ Clean up after phase is done """
        pass

    def state_dict(self):
        """
        State to save down
        """
        return {}

    def banner(self) -> str:
        """ Return banner for the phase """
        return f"|------> PHASE: {self.__class__.__name__} Length: {self.number_of_epochs}"


class EmptyTrainPhase(TrainPhase):
    """ A train phase that is a simple call, without any training """

    @property
    def number_of_epochs(self) -> int:
        """ How many epochs does this phase take """
        return 0

    def execute_epoch(self, epoch_info, learner):
        """ Prepare the phase for learning """
        pass

    def epoch_info(self, training_info: TrainingInfo, global_idx: int, local_idx: int) -> EpochInfo:
        """ Create Epoch info """
        return EpochInfo(training_info, global_epoch_idx=global_idx, local_epoch_idx=local_idx, batches_per_epoch=0)



import torch.nn as nn


ACTIVATION_DICT = {
    'relu': nn.ReLU,
    'tanh': nn.Tanh,
    'elu': nn.ELU,
    'leaky_relu': nn.LeakyReLU
}

NORMALIZATION_DICT = {
    'layer': nn.LayerNorm,
    'layer-noscale': lambda normalized_shape: nn.LayerNorm(normalized_shape, elementwise_affine=False),
    'batch1d': nn.BatchNorm1d,
    'batch2d': nn.BatchNorm2d
}


def activation(name):
    """ Return activation block corresponding given name """
    return ACTIVATION_DICT[name]


def normalization(name):
    """ Return activation block corresponding given name """
    return NORMALIZATION_DICT[name]


def convolution_size_equation(size, filter_size, padding, stride):
    """ Output size of convolutional layer """
    return (size - filter_size + 2 * padding) // stride + 1


def convolutional_layer_series(initial_size, layer_sequence):
    """ Execute a series of convolutional layer transformations to the size number """
    size = initial_size

    for filter_size, padding, stride in layer_sequence:
        size = convolution_size_equation(size, filter_size, padding, stride)

    return size



import gym
import os.path

from gym.envs.registration import EnvSpec

from vel.rl.api.base import EnvFactory
from vel.openai.baselines import logger
from vel.openai.baselines.bench import Monitor
from vel.rl.env.wrappers.env_normalize import EnvNormalize


DEFAULT_SETTINGS = {
    'default': {
        'monitor': False,
        'allow_early_resets': False,
        'normalize_observations': False,
        'normalize_returns': False,
    },
    'raw': {
        'monitor': False,
        'allow_early_resets': True,
        'normalize_observations': False,
        'normalize_returns': False,
    }
}


def env_maker(environment_id, seed, serial_id, monitor=False, allow_early_resets=False, normalize_observations=False,
              normalize_returns=False, normalize_gamma=0.99):
    """ Create a relatively raw atari environment """
    env = gym.make(environment_id)
    env.seed(seed + serial_id)

    # Monitoring the env
    if monitor:
        logdir = logger.get_dir() and os.path.join(logger.get_dir(), str(serial_id))
    else:
        logdir = None

    env = Monitor(env, logdir, allow_early_resets=allow_early_resets)

    if normalize_observations or normalize_returns:
        env = EnvNormalize(
            env,
            normalize_observations=normalize_observations,
            normalize_returns=normalize_returns,
            gamma=normalize_gamma
        )

    return env


class MujocoEnv(EnvFactory):
    """ Atari game environment wrapped in the same way as Deep Mind and OpenAI baselines """
    def __init__(self, envname, env_settings=None, normalize_observations=False, normalize_returns=False):
        self.envname = envname

        env_settings = env_settings if env_settings is not None else {}
        env_keys = set(DEFAULT_SETTINGS.keys()).union(set(env_settings.keys()))

        self.presets = {}

        for key in env_keys:
            preset_settings = DEFAULT_SETTINGS.get(key, {}).copy()

            if normalize_observations:
                preset_settings['normalize_observations'] = True

            if normalize_returns:
                preset_settings['normalize_returns'] = True

            if key in env_settings:
                preset_settings.update(env_settings[key])

            self.presets[key] = preset_settings

    def specification(self) -> EnvSpec:
        """ Return environment specification """
        return gym.spec(self.envname)

    def get_preset(self, preset_key='default') -> dict:
        """ Get env settings for given preset """
        return self.presets[preset_key]

    def instantiate(self, seed=0, serial_id=0, preset='default', extra_args=None) -> gym.Env:
        """ Make a single environment compatible with the experiments """
        settings = self.get_preset(preset)
        return env_maker(self.envname, seed, serial_id, **settings)


def create(game, env_settings=None, normalize_observations=False, normalize_returns=False):
    return MujocoEnv(
        envname=game,
        env_settings=env_settings,
        normalize_observations=normalize_observations,
        normalize_returns=normalize_returns
    )



import gym
import os.path

from gym.envs.registration import EnvSpec


from vel.openai.baselines import logger
from vel.openai.baselines.bench import Monitor
from vel.openai.baselines.common.atari_wrappers import (
    NoopResetEnv, MaxAndSkipEnv, FireResetEnv, EpisodicLifeEnv, WarpFrame, ClipRewardEnv,
    ScaledFloatFrame, FrameStack, FireEpisodicLifeEnv
)

from vel.rl.api.base import EnvFactory
from vel.rl.env.wrappers.clip_episode_length import ClipEpisodeLengthWrapper


DEFAULT_SETTINGS = {
    'default': {
        'disable_reward_clipping': False,
        'disable_episodic_life': False,
        'monitor': False,
        'allow_early_resets': False,
        'scale_float_frames': False,
        'max_episode_frames': 10000,
        'frame_stack': None
    },
    'raw': {
        'disable_reward_clipping': False,
        'disable_episodic_life': True,
        'monitor': False,
        'allow_early_resets': True,
        'scale_float_frames': False,
        'max_episode_frames': 10000,
        'frame_stack': None
    },
}


def env_maker(environment_id):
    """ Create a relatively raw atari environment """
    env = gym.make(environment_id)
    assert 'NoFrameskip' in env.spec.id

    # Wait for between 1 and 30 rounds doing nothing on start
    env = NoopResetEnv(env, noop_max=30)

    # Do the same action for k steps. Return max of last 2 frames. Return sum of rewards
    env = MaxAndSkipEnv(env, skip=4)

    return env


def wrapped_env_maker(environment_id, seed, serial_id, disable_reward_clipping=False, disable_episodic_life=False,
                      monitor=False, allow_early_resets=False, scale_float_frames=False,
                      max_episode_frames=10000, frame_stack=None):
    """ Wrap atari environment so that it's nicer to learn RL algorithms """
    env = env_maker(environment_id)
    env.seed(seed + serial_id)

    if max_episode_frames is not None:
        env = ClipEpisodeLengthWrapper(env, max_episode_length=max_episode_frames)

    # Monitoring the env
    if monitor:
        logdir = logger.get_dir() and os.path.join(logger.get_dir(), str(serial_id))
    else:
        logdir = None

    env = Monitor(env, logdir, allow_early_resets=allow_early_resets)

    if not disable_episodic_life:
        # Make end-of-life == end-of-episode, but only reset on true game over.
        # Done by DeepMind for the DQN and co. since it helps value estimation.
        env = EpisodicLifeEnv(env)

    if 'FIRE' in env.unwrapped.get_action_meanings():
        # Take action on reset for environments that are fixed until firing.
        if disable_episodic_life:
            env = FireEpisodicLifeEnv(env)
        else:
            env = FireResetEnv(env)

    # Warp frames to 84x84 as done in the Nature paper and later work.
    env = WarpFrame(env)

    if scale_float_frames:
        env = ScaledFloatFrame(env)

    if not disable_reward_clipping:
        # Bin reward to {+1, 0, -1} by its sign.
        env = ClipRewardEnv(env)

    if frame_stack is not None:
        env = FrameStack(env, frame_stack)

    return env


class ClassicAtariEnv(EnvFactory):
    """ Atari game environment wrapped in the same way as Deep Mind and OpenAI baselines """
    def __init__(self, envname, env_settings=None):
        self.envname = envname

        env_settings = env_settings if env_settings is not None else {}
        env_keys = set(DEFAULT_SETTINGS.keys()).union(set(env_settings.keys()))

        self.presets = {}

        for key in env_keys:
            self.presets[key] = env_settings.get(key, {})

    def specification(self) -> EnvSpec:
        """ Return environment specification """
        return gym.spec(self.envname)

    def get_preset(self, preset_key='default') -> dict:
        """ Get env settings for given preset """
        current_settings = DEFAULT_SETTINGS.get(preset_key, {}).copy()

        if 'all' in self.presets:
            current_settings.update(self.presets['all'])

        # Key must be present in presets
        current_settings.update(self.presets[preset_key])

        return current_settings

    def instantiate(self, seed=0, serial_id=0, preset='default', extra_args=None) -> gym.Env:
        """ Make a single environment compatible with the experiments """
        settings = self.get_preset(preset)
        return wrapped_env_maker(self.envname, seed, serial_id, **settings)


def create(game, env_settings=None):
    return ClassicAtariEnv(game, env_settings)



import torch.optim

import vel.util.module_util as mu

from vel.api.base import OptimizerFactory, Model


class SgdFactory(OptimizerFactory):
    """ SGD optimizer factory """

    def __init__(self, lr, momentum=0, dampening=0, weight_decay=0, nesterov=False, layer_groups: bool=False):
        self.lr = lr
        self.momentum = momentum
        self.dampening = dampening
        self.weight_decay = weight_decay
        self.nesterov = nesterov
        self.layer_groups = layer_groups

    def instantiate(self, model: Model) -> torch.optim.SGD:
        if self.layer_groups:
            parameters = mu.to_parameter_groups(model.get_layer_groups())
        else:
            parameters = filter(lambda p: p.requires_grad, model.parameters())

        return torch.optim.SGD(
            parameters,
            lr=self.lr, momentum=self.momentum, dampening=self.dampening, weight_decay=self.weight_decay,
            nesterov=self.nesterov
        )


def create(lr, weight_decay=0, momentum=0, layer_groups=False):
    """ Return an SGD optimizer """
    return SgdFactory(lr=lr, weight_decay=weight_decay, momentum=momentum, layer_groups=layer_groups)



import typing
import gym

from vel.rl.api.rollout import Rollout
from vel.api import BatchInfo
from vel.api.base import Model
from vel.openai.baselines.common.vec_env import VecEnv


class EnvRollerBase:
    """ Class generating environment rollouts """

    @property
    def environment(self) -> typing.Union[gym.Env, VecEnv]:
        """ Reference to environment being evaluated """
        raise NotImplementedError

    def rollout(self, batch_info: BatchInfo, model: Model) -> Rollout:
        """ Roll-out the environment and return it """
        raise NotImplementedError

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        return []


# noinspection PyAbstractClass
class ReplayEnvRollerBase(EnvRollerBase):
    """ Class generating environment rollouts with experience replay """

    def sample(self, batch_info: BatchInfo, model: Model) -> Rollout:
        """ Sample experience from replay buffer and return a batch """
        raise NotImplementedError

    def is_ready_for_sampling(self) -> bool:
        """ If buffer is ready for drawing samples from it (usually checks if there is enough data) """
        raise NotImplementedError

    def update(self, rollout, batch_info):
        """ Perform update of the internal state of the buffer - e.g. for the prioritized replay weights """
        pass


class EnvRollerFactory:
    """ Factory for env rollers """

    def instantiate(self, environment, device, settings) -> EnvRollerBase:
        """ Instantiate env roller """
        raise NotImplementedError


class ReplayEnvRollerFactory(EnvRollerFactory):
    """ Factory for env rollers """

    def instantiate(self, environment, device, settings) -> ReplayEnvRollerBase:
        """ Instantiate env roller """
        raise NotImplementedError



"""
Learning rate finder.
Loosely based on: https://github.com/fastai/fastai/blob/master/fastai/learner.py
"""
import matplotlib.pyplot as plt
import numpy as np
import torch
import tqdm

import vel.util.intepolate as interp

from vel.api import Learner, TrainingInfo, EpochInfo, BatchInfo
from vel.api.metrics.averaging_metric import AveragingNamedMetric


class LrFindCommand:
    """Helps you find an optimal learning rate for a model.

     It uses the technique developed in the 2015 paper
     `Cyclical Learning Rates for Training Neural Networks`, where
     we simply keep increasing the learning rate from a very small value,
     until the loss starts decreasing.

    Args:
        start_lr (float/numpy array) : Passing in a numpy array allows you
            to specify learning rates for a learner's layer_groups
        end_lr (float) : The maximum learning rate to try.
        wds (iterable/float)

    Examples:
        As training moves us closer to the optimal weights for a model,
        the optimal learning rate will be smaller. We can take advantage of
        that knowledge and provide lr_find() with a starting learning rate
        1000x smaller than the model's current learning rate as such:

        >> learn.lr_find(lr/1000)

        >> lrs = np.array([ 1e-4, 1e-3, 1e-2 ])
        >> learn.lr_find(lrs / 1000)

    Notes:
        lr_find() may finish before going through each batch of examples if
        the loss decreases enough.

    .. _Cyclical Learning Rates for Training Neural Networks:
        http://arxiv.org/abs/1506.01186

    """
    def __init__(self, model_config, model, source, optimizer_factory, start_lr=1e-5, end_lr=10, num_it=100,
                 interpolation='logscale', freeze=False, stop_dv=True, divergence_threshold=4.0, metric='loss'):
        # Mandatory pieces
        self.model = model
        self.source = source
        self.optimizer_factory = optimizer_factory
        self.model_config = model_config
        # Settings
        self.start_lr = start_lr
        self.end_lr = end_lr
        self.num_it = num_it
        self.interpolation = interpolation
        self.freeze = freeze
        self.stop_dv = stop_dv
        self.divergence_threshold = divergence_threshold
        self.metric = metric

    def run(self):
        """ Run the command with supplied configuration """
        device = torch.device(self.model_config.device)
        learner = Learner(device, self.model.instantiate())

        lr_schedule = interp.interpolate_series(self.start_lr, self.end_lr, self.num_it, self.interpolation)

        if self.freeze:
            learner.model.freeze()

        # Optimizer shoudl be created after freeze
        optimizer = self.optimizer_factory.instantiate(learner.model)

        iterator = iter(self.source.train_loader())

        # Metrics to track through this training
        metrics = learner.metrics() + [AveragingNamedMetric("lr")]

        learner.train()

        best_value = None

        training_info = TrainingInfo(start_epoch_idx=0, metrics=metrics)

        # Treat it all as one epoch
        epoch_info = EpochInfo(
            training_info, global_epoch_idx=1, batches_per_epoch=1, optimizer=optimizer
        )

        for iteration_idx, lr in enumerate(tqdm.tqdm(lr_schedule)):
            batch_info = BatchInfo(epoch_info, iteration_idx)

            # First, set the learning rate, the same for each parameter group
            for param_group in optimizer.param_groups:
                param_group['lr'] = lr

            try:
                data, target = next(iterator)
            except StopIteration:
                iterator = iter(self.source.train_loader())
                data, target = next(iterator)

            learner.train_batch(batch_info, data, target)

            batch_info['lr'] = lr

            # METRIC RECORDING PART
            epoch_info.result_accumulator.calculate(batch_info)

            current_value = epoch_info.result_accumulator.intermediate_value(self.metric)

            final_metrics = {'epoch_idx': iteration_idx, self.metric: current_value, 'lr': lr}

            if best_value is None or current_value < best_value:
                best_value = current_value

            # Stop on divergence
            if self.stop_dv and (np.isnan(current_value) or current_value > best_value * self.divergence_threshold):
                break

            training_info.history.add(final_metrics)

        frame = training_info.history.frame()

        fig, ax = plt.subplots(1, 2)

        ax[0].plot(frame.index, frame.lr)
        ax[0].set_title("LR Schedule")
        ax[0].set_xlabel("Num iterations")
        ax[0].set_ylabel("Learning rate")

        if self.interpolation == 'logscale':
            ax[0].set_yscale("log", nonposy='clip')

        ax[1].plot(frame.lr, frame[self.metric], label=self.metric)
        # ax[1].plot(frame.lr, frame[self.metric].ewm(com=20).mean(), label=self.metric + ' smooth')
        ax[1].set_title(self.metric)
        ax[1].set_xlabel("Learning rate")
        ax[1].set_ylabel(self.metric)
        # ax[1].legend()

        if self.interpolation == 'logscale':
            ax[1].set_xscale("log", nonposx='clip')

        plt.show()


def create(model_config, model, source, optimizer, start_lr=1e-5, end_lr=10, iterations=100, freeze=False,
           interpolation='logscale', stop_dv=True, divergence_threshold=4.0, metric='loss'):
    """ Create a learning rate finder """
    return LrFindCommand(
        model_config=model_config,
        model=model,
        source=source,
        optimizer_factory=optimizer,
        start_lr=start_lr,
        end_lr=end_lr,
        num_it=iterations,
        interpolation=interpolation,
        freeze=freeze,
        stop_dv=stop_dv,
        divergence_threshold=divergence_threshold,
        metric=metric
    )



import gym
import numpy as np
import random

from .deque_backend import DequeBufferBackend


# Segment tree implementation taken from https://github.com/Kaixhin/Rainbow/blob/master/memory.py
class SegmentTree:
    """
    Segment tree data structure where parent node values are sum/max of children node values
    """
    def __init__(self, size):
        self.index = 0
        self.size = size
        self.full = False  # Used to track actual capacity
        self.sum_tree = [0] * (2 * size - 1)  # Initialise fixed size tree with all (priority) zeros
        self.max = 1  # Initial max value to return (1 = 1^ω)

    # Propagates value up tree given a tree index
    def _propagate(self, index, value):
        parent = (index - 1) // 2
        left, right = 2 * parent + 1, 2 * parent + 2
        self.sum_tree[parent] = self.sum_tree[left] + self.sum_tree[right]
        if parent != 0:
            self._propagate(parent, value)

    # Updates value given a tree index
    def update(self, index, value):
        self.sum_tree[index] = value  # Set new value
        self._propagate(index, value)  # Propagate value
        self.max = max(value, self.max)

    # def append(self, data, value):
    def append(self, value):
        # self.data[self.index] = data  # Store data in underlying data structure
        self.update(self.index + self.size - 1, value)  # Update tree
        self.index = (self.index + 1) % self.size  # Update index
        self.full = self.full or self.index == 0  # Save when capacity reached
        self.max = max(value, self.max)

    # Searches for the location of a value in sum tree
    def _retrieve(self, index, value):
        left, right = 2 * index + 1, 2 * index + 2
        if left >= len(self.sum_tree):
            return index
        elif value <= self.sum_tree[left]:
            return self._retrieve(left, value)
        else:
            return self._retrieve(right, value - self.sum_tree[left])

    # Searches for a value in sum tree and returns value, data index and tree index
    def find(self, value):
        index = self._retrieve(0, value)  # Search for index of item from root
        data_index = index - self.size + 1
        return self.sum_tree[index], data_index, index  # Return value, data index, tree index

    def tree_index_for_index(self, index):
        return index + self.size - 1

    def total(self):
        return self.sum_tree[0]


class PrioritizedReplayBackend:
    """ Backend behind the prioritized replay buffer """
    def __init__(self, buffer_capacity: int, observation_space: gym.Space, action_space: gym.Space, extra_data=None):
        self.deque = DequeBufferBackend(buffer_capacity, observation_space, action_space, extra_data=extra_data)
        self.segment_tree = SegmentTree(buffer_capacity)

    def store_transition(self, frame, action, reward, done, extra_info=None):
        """ Store given transition in the backend """
        self.deque.store_transition(frame, action, reward, done, extra_info=extra_info)

        # We add element with max priority to the tree
        self.segment_tree.append(self.segment_tree.max)

    def get_frame(self, idx, history):
        """ Return frame from the buffer """
        return self.deque.get_frame(idx, history)

    def get_frame_with_future(self, idx, history):
        """ Return frame from the buffer together with the next frame """
        return self.deque.get_frame_with_future(idx, history)

    def get_batch(self, indexes, history):
        """ Return batch of frames for given indexes """
        return self.deque.get_batch(indexes, history)

    def update_priority(self, tree_idx, priority):
        """ Update priorities of the elements in the tree """
        self.segment_tree.update(tree_idx, priority)

    def sample_batch_prioritized(self, batch_size, history):
        """ Return indexes of the next sample in from prioritized distribution """
        p_total = self.segment_tree.total()
        segment = p_total / batch_size

        # Get batch of valid samples
        batch = [self._get_sample_from_segment(segment, i, history) for i in range(batch_size)]
        probs, idxs, tree_idxs = zip(*batch)
        return probs, np.array(idxs), tree_idxs

    def _get_sample_from_segment(self, segment, i, history):
        valid = False

        prob = None
        idx = None
        tree_idx = None

        while not valid:
            # Uniformly sample an element from within a segment
            sample = random.uniform(i * segment, (i + 1) * segment)

            # Retrieve sample from tree with un-normalised probability
            prob, idx, tree_idx = self.segment_tree.find(sample)

            # Resample if transition straddled current index or probablity 0
            if (self.segment_tree.index - idx) % self.segment_tree.size > 1 and \
                    (idx - self.segment_tree.index) % self.segment_tree.size >= history and prob != 0:
                valid = True  # Note that conditions are valid but extra conservative around buffer index 0

        return prob, idx, tree_idx

    @property
    def current_size(self):
        """ Return current size of the replay buffer """
        return self.deque.current_size

    @property
    def current_idx(self):
        """ Return current index """
        return self.deque.current_idx



import torch.optim

from vel.api.base import OptimizerFactory, Model


class RMSpropFactory(OptimizerFactory):
    """ RMSprop optimizer factory """

    def __init__(self, lr=1e-2, alpha=0.99, eps=1e-8, weight_decay=0, momentum=0, centered=False):
        self.lr = lr
        self.alpha = alpha
        self.eps = eps
        self.weight_decay = weight_decay
        self.momentum = momentum
        self.centered = centered

    def instantiate(self, model: Model) -> torch.optim.RMSprop:
        return torch.optim.RMSprop(
            filter(lambda p: p.requires_grad, model.parameters()),
            lr=self.lr, alpha=self.alpha, eps=self.eps,
            weight_decay=self.weight_decay, momentum=self.momentum, centered=self.centered
        )


def create(lr, alpha, momentum=0, weight_decay=0, epsilon=1e-8):
    """ Vel creation function - RMSprop optimizer"""
    return RMSpropFactory(lr=lr, alpha=alpha, momentum=momentum, weight_decay=weight_decay, eps=float(epsilon))



import vel.api.base as base

from vel.api import TrainingInfo, EpochInfo
from vel.api.base import Source


class GenericPhase(base.TrainPhase):
    """ Most generic phase of training """

    def __init__(self, lr, epochs, optimizer_factory):
        self.lr = lr
        self.epochs = epochs
        self.optimizer_factory = optimizer_factory

        self._optimizer_instance = None
        self._source = None

    @property
    def number_of_epochs(self) -> int:
        return self.epochs

    def set_up_phase(self, training_info, model, source: Source):
        """ Prepare the phase for learning """
        self._optimizer_instance = self.optimizer_factory.instantiate(model)
        self._source = source

    def epoch_info(self, training_info: TrainingInfo, global_idx: int, local_idx: int) -> EpochInfo:
        """ Create Epoch info """
        return EpochInfo(
            training_info=training_info,
            global_epoch_idx=global_idx,
            local_epoch_idx=local_idx,
            batches_per_epoch=self._source.train_iterations_per_epoch(),
            optimizer=self._optimizer_instance
        )

    def execute_epoch(self, epoch_info, learner):
        """ Prepare the phase for learning """
        for param_group in epoch_info.optimizer.param_groups:
            param_group['lr'] = self.lr

        epoch_result = learner.run_epoch(epoch_info, self._source)

        return epoch_result


def create(lr, epochs, optimizer):
    """ Vel creation function """
    return GenericPhase(
        lr=lr,
        epochs=epochs,
        optimizer_factory=optimizer,
    )



import numpy as np

import vel.api.data as data


class Normalize(data.Augmentation):
    """ Normalize input mean and standard deviation """

    def __init__(self, mean, std, mode='x', tags=None):
        super().__init__(mode, tags)
        self.mean = np.array(mean, dtype=np.float32)
        self.std = np.array(std, dtype=np.float32)

    def __call__(self, x_data):
        return (x_data - self.mean) / self.std

    def denormalize(self, x_data):
        """ Operation reverse to normalization """
        return x_data * self.std + self.mean


def create(mean, std, mode='x', tags=None):
    return Normalize(mean=mean, std=std, mode=mode, tags=tags)




import torch
import typing

from vel.api import ModelConfig, EpochInfo, TrainingInfo, BatchInfo
from vel.api.base import OptimizerFactory, Storage, Callback
from vel.rl.api.base import ReinforcerFactory
from vel.callbacks.time_tracker import TimeTracker

import vel.openai.baselines.logger as openai_logger


class FrameTracker(Callback):
    """ Aggregate frame count from each batch to a global number """
    def __init__(self, max_frames: typing.Optional[typing.Union[int, float]] = None):
        self.max_frames = max_frames

    def on_initialization(self, training_info: TrainingInfo):
        if self.max_frames is not None:
            training_info['total_frames'] = int(self.max_frames)

        training_info['frames'] = 0

    def on_batch_begin(self, batch_info: BatchInfo):
        if 'total_frames' in batch_info.training_info:
            # Track progress during learning
            batch_info['progress'] = (
                float(batch_info.training_info['frames']) / batch_info.training_info['total_frames']
            )

    def on_batch_end(self, batch_info: BatchInfo):
        batch_info.training_info['frames'] += batch_info['frames']

    def write_state_dict(self, training_info: TrainingInfo, hidden_state_dict: dict):
        hidden_state_dict['frame_tracker/frames'] = training_info['frames']

    def load_state_dict(self, training_info: TrainingInfo, hidden_state_dict: dict):
        training_info['frames'] = hidden_state_dict['frame_tracker/frames']


class RlTrainCommand:
    """ Train a reinforcement learning algorithm by evaluating the environment and """
    def __init__(self, model_config: ModelConfig, reinforcer: ReinforcerFactory,
                 optimizer_factory: OptimizerFactory,
                 storage: Storage, callbacks,
                 total_frames: int, batches_per_epoch: int,
                 scheduler_factory=None, openai_logging=False):
        self.model_config = model_config
        self.reinforcer = reinforcer
        self.optimizer_factory = optimizer_factory
        self.scheduler_factory = scheduler_factory
        self.storage = storage
        self.total_frames = total_frames
        self.batches_per_epoch = batches_per_epoch
        self.callbacks = callbacks if callbacks is not None else []

        self.openai_logging = openai_logging

    def run(self):
        """ Run reinforcement learning algorithm """
        device = torch.device(self.model_config.device)
        # Reinforcer is the learner for the reinforcement learning model
        reinforcer = self.reinforcer.instantiate(device)
        optimizer = self.optimizer_factory.instantiate(reinforcer.model)

        # All callbacks used for learning
        callbacks = self.gather_callbacks(optimizer)
        # Metrics to track through this training
        metrics = reinforcer.metrics()

        training_info = self.resume_training(reinforcer, callbacks, metrics)

        reinforcer.initialize_training(training_info)
        training_info.on_train_begin()

        if training_info.optimizer_initial_state:
            optimizer.load_state_dict(training_info.optimizer_initial_state)

        global_epoch_idx = training_info.start_epoch_idx + 1

        while training_info['frames'] < self.total_frames:
            epoch_info = EpochInfo(
                training_info,
                global_epoch_idx=global_epoch_idx,
                batches_per_epoch=self.batches_per_epoch,
                optimizer=optimizer,
            )

            reinforcer.train_epoch(epoch_info)

            if self.openai_logging:
                self._openai_logging(epoch_info.result)

            self.storage.checkpoint(epoch_info, reinforcer.model)

            global_epoch_idx += 1

        training_info.on_train_end()

        return training_info

    def gather_callbacks(self, optimizer) -> list:
        """ Gather all the callbacks to be used in this training run """
        callbacks = [FrameTracker(self.total_frames), TimeTracker()]

        if self.scheduler_factory is not None:
            callbacks.append(self.scheduler_factory.instantiate(optimizer))

        callbacks.extend(self.callbacks)
        callbacks.extend(self.storage.streaming_callbacks())

        return callbacks

    def resume_training(self, reinforcer, callbacks, metrics) -> TrainingInfo:
        """ Possibly resume training from a saved state from the storage """
        if self.model_config.continue_training:
            start_epoch = self.storage.last_epoch_idx()
        else:
            start_epoch = 0

        training_info = TrainingInfo(
            start_epoch_idx=start_epoch,
            run_name=self.model_config.run_name,
            metrics=metrics, callbacks=callbacks
        )

        if start_epoch == 0:
            self.storage.reset(self.model_config.render_configuration())
            training_info.initialize()
            reinforcer.initialize_training(training_info)
        else:
            self.storage.resume(training_info, reinforcer.model)

        return training_info

    def _openai_logging(self, epoch_result):
        """ Use OpenAI logging facilities for the same type of logging """
        for key in sorted(epoch_result.keys()):
            if key == 'fps':
                # Not super elegant, but I like nicer display of FPS
                openai_logger.record_tabular(key, int(epoch_result[key]))
            else:
                openai_logger.record_tabular(key, epoch_result[key])

        openai_logger.dump_tabular()


def create(model_config, reinforcer, optimizer, storage,
           # Settings:
           total_frames, batches_per_epoch,  callbacks=None, scheduler=None, openai_logging=False):
    """ Create reinforcement learning pipeline """
    from vel.openai.baselines import logger
    logger.configure(dir=model_config.openai_dir())

    return RlTrainCommand(
        model_config=model_config,
        reinforcer=reinforcer,
        optimizer_factory=optimizer,
        scheduler_factory=scheduler,
        storage=storage,
        callbacks=callbacks,
        total_frames=int(float(total_frames)),
        batches_per_epoch=int(batches_per_epoch),
        openai_logging=openai_logging
    )



import os
import pathlib
import re
import torch


from vel.api import ModelConfig, EpochInfo, TrainingInfo
from vel.api.base import Model, Storage
from .strategy.checkpoint_strategy import CheckpointStrategy


class ClassicStorage(Storage):
    """ Model and metric persistence - classic implementation """

    def __init__(self, model_config: ModelConfig, checkpoint_strategy: CheckpointStrategy, backend, streaming=None):
        self.model_config = model_config
        self.backend = backend
        self.streaming = streaming if streaming is not None else []
        self.checkpoint_strategy = checkpoint_strategy

        self.cleaned = False

    def last_epoch_idx(self):
        """ Return last checkpointed epoch idx for given configuration. Returns 0 if no results have been stored """
        return self._persisted_last_epoch()

    def reset(self, configuration: dict) -> None:
        """
        Whenever there was anything stored in the database or not, purge previous state and start
        new training process from scratch.
        """
        self.clean(0)
        self.backend.store_config(configuration)

    def resume(self, train_info: TrainingInfo, model: Model) -> dict:
        """
        Resume learning process and return loaded hidden state dictionary
        """
        last_epoch = train_info.start_epoch_idx

        model.load_state_dict(torch.load(self.checkpoint_filename(last_epoch)))
        hidden_state = torch.load(self.checkpoint_hidden_filename(last_epoch))

        self.checkpoint_strategy.restore(hidden_state)
        train_info.restore(hidden_state)

        return hidden_state

    def get_metrics_frame(self):
        """ Get a frame of metrics from backend """
        return self.backend.get_frame()

    def clean(self, global_epoch_idx):
        """ Clean old checkpoints """
        if self.cleaned:
            return

        self.cleaned = True
        self.backend.clean(global_epoch_idx)

        self._make_sure_dir_exists()

        for x in os.listdir(self.model_config.checkpoint_dir()):
            match = re.match('checkpoint_(\\d+)\\.data', x)

            if match:
                idx = int(match[1])

                if idx > global_epoch_idx:
                    os.remove(os.path.join(self.model_config.checkpoint_dir(), x))

            match = re.match('checkpoint_hidden_(\\d+)\\.data', x)

            if match:
                idx = int(match[1])

                if idx > global_epoch_idx:
                    os.remove(os.path.join(self.model_config.checkpoint_dir(), x))

            match = re.match('checkpoint_best_(\\d+)\\.data', x)

            if match:
                idx = int(match[1])

                if idx > global_epoch_idx:
                    os.remove(os.path.join(self.model_config.checkpoint_dir(), x))

    def checkpoint(self, epoch_info: EpochInfo, model: Model):
        """ When epoch is done, we persist the training state """
        self.clean(epoch_info.global_epoch_idx - 1)

        self._make_sure_dir_exists()

        # Checkpoint latest
        torch.save(model.state_dict(), self.checkpoint_filename(epoch_info.global_epoch_idx))

        hidden_state = epoch_info.state_dict()
        self.checkpoint_strategy.write_state_dict(hidden_state)

        torch.save(hidden_state, self.checkpoint_hidden_filename(epoch_info.global_epoch_idx))

        if epoch_info.global_epoch_idx > 1 and self.checkpoint_strategy.should_delete_previous_checkpoint(
                                                   epoch_info.global_epoch_idx):
            prev_epoch_idx = epoch_info.global_epoch_idx - 1

            os.remove(self.checkpoint_filename(prev_epoch_idx))
            os.remove(self.checkpoint_hidden_filename(prev_epoch_idx))

        if self.checkpoint_strategy.should_store_best_checkpoint(epoch_info.global_epoch_idx, epoch_info.result):
            best_checkpoint_idx = self.checkpoint_strategy.current_best_checkpoint_idx

            if best_checkpoint_idx is not None:
                os.remove(self.checkpoint_best_filename(best_checkpoint_idx))

            torch.save(model.state_dict(), self.checkpoint_best_filename(epoch_info.global_epoch_idx))

            self.checkpoint_strategy.store_best_checkpoint_idx(epoch_info.global_epoch_idx)

        self.backend.store(epoch_info.result)

    def streaming_callbacks(self) -> list:
        """ Lift of callbacks for live streaming results """
        return self.streaming

    ####################################################################################################################
    # Filename helpers
    def checkpoint_filename(self, epoch_idx) -> str:
        """ Return checkpoint filename for this model """
        return self.model_config.checkpoint_dir('checkpoint_{:08}.data'.format(epoch_idx))

    def checkpoint_best_filename(self, epoch_idx) -> str:
        """ Return checkpoint filename for this model - best version """
        return self.model_config.checkpoint_dir('checkpoint_best_{:08}.data'.format(epoch_idx))

    def checkpoint_hidden_filename(self, epoch_idx) -> str:
        """ Return checkpoint filename for this model - hidden state """
        return self.model_config.checkpoint_dir('checkpoint_hidden_{:08}.data'.format(epoch_idx))

    ####################################################################################################################
    # Internal interface
    def _persisted_last_epoch(self) -> int:
        """ Return number of last epoch already calculated """
        epoch_number = 0
        self._make_sure_dir_exists()

        for x in os.listdir(self.model_config.checkpoint_dir()):
            match = re.match('checkpoint_(\\d+)\\.data', x)
            if match:
                idx = int(match[1])

                if idx > epoch_number:
                    epoch_number = idx

        return epoch_number

    def _make_sure_dir_exists(self):
        """ Make sure directory exists """
        filename = self.model_config.checkpoint_dir()
        pathlib.Path(filename).mkdir(parents=True, exist_ok=True)


def create(model_config, backend, checkpoint_strategy, streaming=None):
    """ Vel creation function """
    return ClassicStorage(
        model_config=model_config,
        backend=backend,
        checkpoint_strategy=checkpoint_strategy,
        streaming=streaming
    )



from vel.api.base import Schedule


class ConstantSchedule(Schedule):
    """ Interpolate variable linearly between start value and final value """

    def __init__(self, value):
        self._value = value

    def value(self, progress_indicator):
        """ Interpolate linearly between start and end """
        return self._value


def create(value):
    """ Vel creation function """
    return ConstantSchedule(value)



import gym
import torch
import torch.nn as nn
import typing

from vel.api.base import RnnLinearBackboneModel, RnnModel, ModelFactory
from vel.rl.api import Rollout, Trajectories, Evaluator
from vel.rl.modules.action_head import ActionHead
from vel.rl.modules.value_head import ValueHead


class PolicyGradientRnnEvaluator(Evaluator):
    """ Evaluate recurrent model from initial state """

    def __init__(self, model: 'PolicyGradientRnnModel', rollout: Rollout):
        assert isinstance(rollout, Trajectories), "For an RNN model, we must evaluate trajectories"
        super().__init__(rollout)

        self.model = model

        observation_trajectories = rollout.transition_tensors['observations']
        hidden_state = rollout.rollout_tensors['initial_hidden_state']

        action_accumulator = []
        value_accumulator = []

        # Evaluate recurrent network step by step
        for i in range(observation_trajectories.size(0)):
            action_output, value_output, hidden_state = model(observation_trajectories[i], hidden_state)

            action_accumulator.append(action_output)
            value_accumulator.append(value_output)

        policy_params = torch.cat(action_accumulator, dim=0)
        estimated_values = torch.cat(value_accumulator, dim=0)

        self.provide('model:policy_params', policy_params)
        self.provide('model:estimated_values', estimated_values)

    @Evaluator.provides('model:action:logprobs')
    def model_action_logprobs(self):
        actions = self.get('rollout:actions')
        policy_params = self.get('model:policy_params')
        return self.model.action_head.logprob(actions, policy_params)

    @Evaluator.provides('model:entropy')
    def model_entropy(self):
        policy_params = self.get('model:policy_params')
        return self.model.entropy(policy_params)


class PolicyGradientRnnModel(RnnModel):
    """ For a policy gradient algorithm we need set of custom heads for our model """

    def __init__(self, backbone: RnnLinearBackboneModel, action_space: gym.Space,
                 input_block: typing.Optional[nn.Module]=None):
        super().__init__()

        self.input_block = input_block
        self.backbone = backbone

        self.action_head = ActionHead(
            action_space=action_space,
            input_dim=self.backbone.output_dim
        )
        self.value_head = ValueHead(input_dim=self.backbone.output_dim)

        assert self.backbone.is_recurrent, "Backbone must be a recurrent model"

    @property
    def state_dim(self) -> int:
        """ Dimension of model state """
        return self.backbone.state_dim

    def reset_weights(self):
        """ Initialize properly model weights """
        self.backbone.reset_weights()
        self.action_head.reset_weights()
        self.value_head.reset_weights()

    def forward(self, observations, state):
        """ Calculate model outputs """
        if self.input_block is not None:
            input_data = self.input_block(observations)
        else:
            input_data = observations

        base_output, new_state = self.backbone(input_data, state=state)

        action_output = self.action_head(base_output)
        value_output = self.value_head(base_output)

        return action_output, value_output, new_state

    def step(self, observations, state, argmax_sampling=False):
        """ Select actions based on model's output """
        action_pd_params, value_output, new_state = self(observations, state)
        actions = self.action_head.sample(action_pd_params, argmax_sampling=argmax_sampling)

        # log likelihood of selected action
        logprobs = self.action_head.logprob(actions, action_pd_params)

        return {
            'actions': actions,
            'values': value_output,
            'logprobs': logprobs,
            'state': new_state
        }

    def evaluate(self, rollout: Rollout) -> Evaluator:
        """ Evaluate model on a rollout """
        return PolicyGradientRnnEvaluator(self, rollout)

    def logprob(self, action_sample, policy_params):
        """ Calculate - log(prob) of selected actions """
        return self.action_head.logprob(action_sample, policy_params)

    def value(self, observations, state):
        """ Calculate only value head for given state """
        if self.input_block is not None:
            input_data = self.input_block(observations)
        else:
            input_data = observations

        base_output, new_state = self.backbone(input_data, state)
        value_output = self.value_head(base_output)

        return value_output

    def entropy(self, action_pd_params):
        """ Entropy of a probability distribution """
        return self.action_head.entropy(action_pd_params)


class PolicyGradientRnnModelFactory(ModelFactory):
    """ Factory  class for policy gradient models """
    def __init__(self, backbone: ModelFactory, input_block=None):
        self.backbone = backbone
        self.input_block = input_block

    def instantiate(self, **extra_args):
        """ Instantiate the model """
        backbone = self.backbone.instantiate(**extra_args)

        if self.input_block is None:
            input_block = None
        else:
            input_block = self.input_block.instantiate()

        return PolicyGradientRnnModel(backbone, extra_args['action_space'], input_block)


def create(backbone: ModelFactory, input_block=None):
    """ Vel creation function """
    return PolicyGradientRnnModelFactory(backbone=backbone, input_block=input_block)



import collections
import torch.optim

import vel.util.module_util as mu
from vel.api.base import OptimizerFactory, Model


class AdamFactory(OptimizerFactory):
    """ Adam optimizer factory """

    def __init__(self, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, amsgrad=False, layer_groups=False):
        self.lr = lr
        self.betas = betas
        self.eps = eps
        self.weight_decay = weight_decay
        self.amsgrad = amsgrad
        self.layer_groups = layer_groups

    def instantiate(self, model: Model) -> torch.optim.Adam:
        if self.layer_groups:
            parameters = mu.to_parameter_groups(model.get_layer_groups())

            if isinstance(self.lr, collections.Sequence):
                for idx, lr in enumerate(self.lr):
                    parameters[idx]['lr'] = lr

                default_lr = self.lr[0]
            else:
                default_lr = float(self.lr)

            if isinstance(self.weight_decay, collections.Sequence):
                for idx, weight_decay in enumerate(self.weight_decay):
                    parameters[idx]['weight_decay'] = weight_decay

                default_weight_decay = self.weight_decay[0]
            else:
                default_weight_decay = self.weight_decay

            return torch.optim.Adam(
                parameters,
                lr=default_lr, betas=self.betas, eps=self.eps, weight_decay=default_weight_decay, amsgrad=self.amsgrad
            )
        else:
            parameters = filter(lambda p: p.requires_grad, model.parameters())

            return torch.optim.Adam(
                parameters,
                lr=self.lr, betas=self.betas, eps=self.eps, weight_decay=self.weight_decay, amsgrad=self.amsgrad
            )


def create(lr, weight_decay=0, epsilon=1e-8, layer_groups=False):
    """ Return an ADAM optimizer """
    return AdamFactory(lr=lr, weight_decay=weight_decay, eps=epsilon, layer_groups=layer_groups)



from vel.openai.baselines import logger

from vel.rl.env.classic_atari import ClassicAtariEnv
from vel.rl.vecenv.subproc import SubprocVecEnvWrapper

from secret.secret_acer_setup.acer_train import learn


def main():
    logger.configure()

    vec_env = SubprocVecEnvWrapper(
        ClassicAtariEnv('BeamRiderNoFrameskip-v4'),
        # frame_history=4
    ).instantiate(parallel_envs=12, seed=0)

    learn(
        env=vec_env,
        seed=0,
        total_timesteps=int(1.1e7),
        lrschedule='constant',
        trust_region=False
        # network='cnn'
    )


if __name__ == '__main__':
    main()



import torch


def explained_variance(returns, values):
    """ Calculate how much variance in returns do the values explain """
    exp_var = 1 - torch.var(returns - values) / torch.var(returns)
    return exp_var.item()



import torch
import torch.optim as optim

from vel.rl.commands.rl_train_command import FrameTracker
from vel.rl.metrics import EpisodeRewardMetric
from vel.schedules.linear import LinearSchedule
from vel.schedules.linear_and_constant import LinearAndConstantSchedule
from vel.util.random import set_seed

from vel.rl.env.classic_atari import ClassicAtariEnv
from vel.rl.env.mujoco import MujocoEnv
from vel.rl.vecenv.subproc import SubprocVecEnvWrapper
from vel.rl.vecenv.dummy import DummyVecEnvWrapper

from vel.rl.models.policy_gradient_model import PolicyGradientModelFactory
from vel.rl.models.q_policy_gradient_model import QPolicyGradientModelFactory
from vel.rl.models.q_model import QModelFactory
from vel.rl.models.deterministic_policy_model import DeterministicPolicyModelFactory
from vel.rl.models.policy_gradient_model_separate import PolicyGradientModelSeparateFactory

from vel.rl.models.backbone.nature_cnn import NatureCnnFactory
from vel.rl.models.backbone.mlp import MLPFactory

from vel.rl.reinforcers.on_policy_iteration_reinforcer import (
    OnPolicyIterationReinforcer, OnPolicyIterationReinforcerSettings
)

from vel.rl.reinforcers.buffered_single_off_policy_iteration_reinforcer import (
    BufferedSingleOffPolicyIterationReinforcer, BufferedSingleOffPolicyIterationReinforcerSettings
)

from vel.rl.reinforcers.buffered_mixed_policy_iteration_reinforcer import (
    BufferedMixedPolicyIterationReinforcer, BufferedMixedPolicyIterationReinforcerSettings
)

from vel.rl.algo.dqn import DeepQLearning
from vel.rl.algo.policy_gradient.a2c import A2CPolicyGradient
from vel.rl.algo.policy_gradient.ppo import PpoPolicyGradient
from vel.rl.algo.policy_gradient.trpo import TrpoPolicyGradient
from vel.rl.algo.policy_gradient.acer import AcerPolicyGradient
from vel.rl.algo.policy_gradient.ddpg import DeepDeterministicPolicyGradient

from vel.rl.env_roller.vec.step_env_roller import StepEnvRoller
from vel.rl.env_roller.vec.replay_q_env_roller import ReplayQEnvRoller
from vel.rl.env_roller.single.deque_replay_roller_epsgreedy import DequeReplayRollerEpsGreedy
from vel.rl.env_roller.single.prioritized_replay_roller_epsgreedy import PrioritizedReplayRollerEpsGreedy
from vel.rl.env_roller.single.deque_replay_roller_ou_noise import DequeReplayRollerOuNoise

from vel.api.info import TrainingInfo, EpochInfo


def test_a2c_breakout():
    """
    Simple 1 iteration of a2c breakout
    """
    device = torch.device('cpu')
    seed = 1001

    # Set random seed in python std lib, numpy and pytorch
    set_seed(seed)

    # Create 16 environments evaluated in parallel in sub processess with all usual DeepMind wrappers
    # These are just helper functions for that
    vec_env = SubprocVecEnvWrapper(
        ClassicAtariEnv('BreakoutNoFrameskip-v4'), frame_history=4
    ).instantiate(parallel_envs=16, seed=seed)

    # Again, use a helper to create a model
    # But because model is owned by the reinforcer, model should not be accessed using this variable
    # but from reinforcer.model property
    model = PolicyGradientModelFactory(
        backbone=NatureCnnFactory(input_width=84, input_height=84, input_channels=4)
    ).instantiate(action_space=vec_env.action_space)

    # Reinforcer - an object managing the learning process
    reinforcer = OnPolicyIterationReinforcer(
        device=device,
        settings=OnPolicyIterationReinforcerSettings(
            discount_factor=0.99,
            batch_size=256,
        ),
        model=model,
        algo=A2CPolicyGradient(
            entropy_coefficient=0.01,
            value_coefficient=0.5,
            max_grad_norm=0.5
        ),
        env_roller=StepEnvRoller(
            environment=vec_env,
            device=device,
            number_of_steps=5,
            discount_factor=0.99,
        )
    )

    # Model optimizer
    optimizer = optim.RMSprop(reinforcer.model.parameters(), lr=7.0e-4, eps=1e-3)

    # Overall information store for training information
    training_info = TrainingInfo(
        metrics=[
            EpisodeRewardMetric('episode_rewards'),  # Calculate average reward from episode
        ],
        callbacks=[]  # Print live metrics every epoch to standard output
    )

    # A bit of training initialization bookkeeping...
    training_info.initialize()
    reinforcer.initialize_training(training_info)
    training_info.on_train_begin()

    # Let's make 100 batches per epoch to average metrics nicely
    num_epochs = 1

    # Normal handrolled training loop
    for i in range(1, num_epochs+1):
        epoch_info = EpochInfo(
            training_info=training_info,
            global_epoch_idx=i,
            batches_per_epoch=1,
            optimizer=optimizer
        )

        reinforcer.train_epoch(epoch_info, interactive=False)

    training_info.on_train_end()


def test_ppo_breakout():
    """
    Simple 1 iteration of ppo breakout
    """
    device = torch.device('cpu')
    seed = 1001

    # Set random seed in python std lib, numpy and pytorch
    set_seed(seed)

    # Create 16 environments evaluated in parallel in sub processess with all usual DeepMind wrappers
    # These are just helper functions for that
    vec_env = SubprocVecEnvWrapper(
        ClassicAtariEnv('BreakoutNoFrameskip-v4'), frame_history=4
    ).instantiate(parallel_envs=8, seed=seed)

    # Again, use a helper to create a model
    # But because model is owned by the reinforcer, model should not be accessed using this variable
    # but from reinforcer.model property
    model = PolicyGradientModelFactory(
        backbone=NatureCnnFactory(input_width=84, input_height=84, input_channels=4)
    ).instantiate(action_space=vec_env.action_space)

    # Reinforcer - an object managing the learning process
    reinforcer = OnPolicyIterationReinforcer(
        device=device,
        settings=OnPolicyIterationReinforcerSettings(
            discount_factor=0.99,
            batch_size=4,
            experience_replay=2,
        ),
        model=model,
        algo=PpoPolicyGradient(
            entropy_coefficient=0.01,
            value_coefficient=0.5,
            max_grad_norm=0.5,
            cliprange=LinearSchedule(0.1, 0.0),
            normalize_advantage=True
        ),
        env_roller=StepEnvRoller(
            environment=vec_env,
            device=device,
            number_of_steps=12,
            discount_factor=0.99,
        )
    )

    # Model optimizer
    # optimizer = optim.RMSprop(reinforcer.model.parameters(), lr=7.0e-4, eps=1e-3)
    optimizer = optim.Adam(reinforcer.model.parameters(), lr=2.5e-4, eps=1e-5)

    # Overall information store for training information
    training_info = TrainingInfo(
        metrics=[
            EpisodeRewardMetric('episode_rewards'),  # Calculate average reward from episode
        ],
        callbacks=[
            FrameTracker(100_000)
        ]  # Print live metrics every epoch to standard output
    )

    # A bit of training initialization bookkeeping...
    training_info.initialize()
    reinforcer.initialize_training(training_info)
    training_info.on_train_begin()

    # Let's make 100 batches per epoch to average metrics nicely
    num_epochs = 1

    # Normal handrolled training loop
    for i in range(1, num_epochs+1):
        epoch_info = EpochInfo(
            training_info=training_info,
            global_epoch_idx=i,
            batches_per_epoch=1,
            optimizer=optimizer
        )

        reinforcer.train_epoch(epoch_info, interactive=False)

    training_info.on_train_end()


def test_dqn_breakout():
    """
    Simple 1 iteration of DQN breakout
    """
    device = torch.device('cpu')
    seed = 1001

    # Set random seed in python std lib, numpy and pytorch
    set_seed(seed)

    # Only single environment for DQN
    env = ClassicAtariEnv('BreakoutNoFrameskip-v4').instantiate(seed=seed)

    # Again, use a helper to create a model
    # But because model is owned by the reinforcer, model should not be accessed using this variable
    # but from reinforcer.model property
    model_factory = QModelFactory(
        backbone=NatureCnnFactory(input_width=84, input_height=84, input_channels=4)
    )

    # Reinforcer - an object managing the learning process
    reinforcer = BufferedSingleOffPolicyIterationReinforcer(
        device=device,
        settings=BufferedSingleOffPolicyIterationReinforcerSettings(
            batch_rollout_rounds=4,
            batch_training_rounds=1,
            batch_size=32,
            discount_factor=0.99
        ),
        environment=env,
        algo=DeepQLearning(
            model_factory=model_factory,
            double_dqn=False,
            target_update_frequency=10_000,
            max_grad_norm=0.5
        ),
        model=model_factory.instantiate(action_space=env.action_space),
        env_roller=DequeReplayRollerEpsGreedy(
            environment=env,
            device=device,
            epsilon_schedule=LinearAndConstantSchedule(initial_value=1.0, final_value=0.1, end_of_interpolation=0.1),
            batch_size=32,
            buffer_capacity=100,
            buffer_initial_size=100,
            frame_stack=4
        )
    )

    # Model optimizer
    optimizer = optim.RMSprop(reinforcer.model.parameters(), lr=2.5e-4, alpha=0.95, momentum=0.95, eps=1e-3)

    # Overall information store for training information
    training_info = TrainingInfo(
        metrics=[
            EpisodeRewardMetric('episode_rewards'),  # Calculate average reward from episode
        ],
        callbacks=[
            FrameTracker(100_000)
        ]  # Print live metrics every epoch to standard output
    )

    # A bit of training initialization bookkeeping...
    training_info.initialize()
    reinforcer.initialize_training(training_info)
    training_info.on_train_begin()

    # Let's make 100 batches per epoch to average metrics nicely
    num_epochs = 1

    # Normal handrolled training loop
    for i in range(1, num_epochs+1):
        epoch_info = EpochInfo(
            training_info=training_info,
            global_epoch_idx=i,
            batches_per_epoch=1,
            optimizer=optimizer
        )

        reinforcer.train_epoch(epoch_info, interactive=False)

    training_info.on_train_end()


def test_prioritized_dqn_breakout():
    """
    Simple 1 iteration of DQN prioritized replay breakout
    """
    device = torch.device('cpu')
    seed = 1001

    # Set random seed in python std lib, numpy and pytorch
    set_seed(seed)

    # Only single environment for DQN
    env = ClassicAtariEnv('BreakoutNoFrameskip-v4').instantiate(seed=seed)

    # Again, use a helper to create a model
    # But because model is owned by the reinforcer, model should not be accessed using this variable
    # but from reinforcer.model property
    model_factory = QModelFactory(
        backbone=NatureCnnFactory(input_width=84, input_height=84, input_channels=4)
    )

    # Reinforcer - an object managing the learning process
    reinforcer = BufferedSingleOffPolicyIterationReinforcer(
        device=device,
        settings=BufferedSingleOffPolicyIterationReinforcerSettings(
            batch_rollout_rounds=4,
            batch_training_rounds=1,
            batch_size=32,
            discount_factor=0.99
        ),
        environment=env,
        algo=DeepQLearning(
            model_factory=model_factory,
            double_dqn=False,
            target_update_frequency=10_000,
            max_grad_norm=0.5
        ),
        model=model_factory.instantiate(action_space=env.action_space),
        env_roller=PrioritizedReplayRollerEpsGreedy(
            environment=env,
            device=device,
            epsilon_schedule=LinearAndConstantSchedule(initial_value=1.0, final_value=0.1, end_of_interpolation=0.1),
            batch_size=8,
            buffer_capacity=100,
            priority_epsilon=1.0e-6,
            buffer_initial_size=100,
            frame_stack=4,
            priority_exponent=0.6,
            priority_weight=LinearSchedule(
                initial_value=0.4,
                final_value=1.0
            ),
        ),
    )

    # Model optimizer
    optimizer = optim.RMSprop(reinforcer.model.parameters(), lr=2.5e-4, alpha=0.95, momentum=0.95, eps=1e-3)

    # Overall information store for training information
    training_info = TrainingInfo(
        metrics=[
            EpisodeRewardMetric('episode_rewards'),  # Calculate average reward from episode
        ],
        callbacks=[
            FrameTracker(100_000)
        ]  # Print live metrics every epoch to standard output
    )

    # A bit of training initialization bookkeeping...
    training_info.initialize()
    reinforcer.initialize_training(training_info)
    training_info.on_train_begin()

    # Let's make 100 batches per epoch to average metrics nicely
    num_epochs = 1

    # Normal handrolled training loop
    for i in range(1, num_epochs+1):
        epoch_info = EpochInfo(
            training_info=training_info,
            global_epoch_idx=i,
            batches_per_epoch=1,
            optimizer=optimizer
        )

        reinforcer.train_epoch(epoch_info, interactive=False)

    training_info.on_train_end()


def test_ddpg_bipedal_walker():
    """
    1 iteration of DDPG bipedal walker environment
    """
    device = torch.device('cpu')
    seed = 1001

    # Set random seed in python std lib, numpy and pytorch
    set_seed(seed)

    # Only single environment for DDPG
    env = MujocoEnv('BipedalWalker-v2').instantiate(seed=seed)

    # Again, use a helper to create a model
    # But because model is owned by the reinforcer, model should not be accessed using this variable
    # but from reinforcer.model property
    model_factory = DeterministicPolicyModelFactory(
        policy_backbone=MLPFactory(input_length=24, hidden_layers=[64, 64], normalization='layer'),
        value_backbone=MLPFactory(input_length=28, hidden_layers=[64, 64], normalization='layer')
    )

    # Reinforcer - an object managing the learning process
    reinforcer = BufferedSingleOffPolicyIterationReinforcer(
        device=device,
        settings=BufferedSingleOffPolicyIterationReinforcerSettings(
            batch_rollout_rounds=4,
            batch_training_rounds=1,
            batch_size=32,
            discount_factor=0.99
        ),
        environment=env,
        algo=DeepDeterministicPolicyGradient(
            model_factory=model_factory,
            tau=0.01,
            max_grad_norm=0.5
        ),
        model=model_factory.instantiate(action_space=env.action_space),
        env_roller=DequeReplayRollerOuNoise(
            environment=env,
            device=device,
            batch_size=32,
            buffer_capacity=100,
            buffer_initial_size=100,
            noise_std_dev=0.2,
            normalize_observations=True,
            discount_factor=0.99
        )
    )

    # Model optimizer
    optimizer = optim.Adam(reinforcer.model.parameters(), lr=2.5e-4, eps=1e-4)

    # Overall information store for training information
    training_info = TrainingInfo(
        metrics=[
            EpisodeRewardMetric('episode_rewards'),  # Calculate average reward from episode
        ],
        callbacks=[
            FrameTracker(100_000)
        ]  # Print live metrics every epoch to standard output
    )

    # A bit of training initialization bookkeeping...
    training_info.initialize()
    reinforcer.initialize_training(training_info)
    training_info.on_train_begin()

    # Let's make 100 batches per epoch to average metrics nicely
    num_epochs = 1

    # Normal handrolled training loop
    for i in range(1, num_epochs+1):
        epoch_info = EpochInfo(
            training_info=training_info,
            global_epoch_idx=i,
            batches_per_epoch=1,
            optimizer=optimizer
        )

        reinforcer.train_epoch(epoch_info, interactive=False)

    training_info.on_train_end()


def test_trpo_bipedal_walker():
    """
    1 iteration of TRPO on bipedal walker
    """
    device = torch.device('cpu')
    seed = 1001

    # Set random seed in python std lib, numpy and pytorch
    set_seed(seed)

    vec_env = DummyVecEnvWrapper(
        MujocoEnv('BipedalWalker-v2'), normalize=True
    ).instantiate(parallel_envs=8, seed=seed)

    # Again, use a helper to create a model
    # But because model is owned by the reinforcer, model should not be accessed using this variable
    # but from reinforcer.model property
    model_factory = PolicyGradientModelSeparateFactory(
        policy_backbone=MLPFactory(input_length=24, hidden_layers=[32, 32]),
        value_backbone=MLPFactory(input_length=24, hidden_layers=[32])
    )

    # Reinforcer - an object managing the learning process
    reinforcer = OnPolicyIterationReinforcer(
        device=device,
        settings=OnPolicyIterationReinforcerSettings(
            discount_factor=0.99,
        ),
        model=model_factory.instantiate(action_space=vec_env.action_space),
        algo=TrpoPolicyGradient(
            max_kl=0.01,
            cg_iters=10,
            line_search_iters=10,
            improvement_acceptance_ratio=0.1,
            cg_damping=0.1,
            vf_iters=5,
            entropy_coef=0.0,
            max_grad_norm=0.5,
        ),
        env_roller=StepEnvRoller(
            environment=vec_env,
            device=device,
            number_of_steps=12,
            discount_factor=0.99,
        )
    )

    # Model optimizer
    optimizer = optim.Adam(reinforcer.model.parameters(), lr=1.0e-3, eps=1e-4)

    # Overall information store for training information
    training_info = TrainingInfo(
        metrics=[
            EpisodeRewardMetric('episode_rewards'),  # Calculate average reward from episode
        ],
        callbacks=[
            FrameTracker(100_000)
        ]  # Print live metrics every epoch to standard output
    )

    # A bit of training initialization bookkeeping...
    training_info.initialize()
    reinforcer.initialize_training(training_info)
    training_info.on_train_begin()

    # Let's make 100 batches per epoch to average metrics nicely
    num_epochs = 1

    # Normal handrolled training loop
    for i in range(1, num_epochs+1):
        epoch_info = EpochInfo(
            training_info=training_info,
            global_epoch_idx=i,
            batches_per_epoch=1,
            optimizer=optimizer
        )

        reinforcer.train_epoch(epoch_info, interactive=False)

    training_info.on_train_end()


def test_acer_breakout():
    """
    1 iteration of ACER on breakout environment
    """
    device = torch.device('cpu')
    seed = 1001

    # Set random seed in python std lib, numpy and pytorch
    set_seed(seed)

    # Create 16 environments evaluated in parallel in sub processess with all usual DeepMind wrappers
    # These are just helper functions for that
    vec_env = SubprocVecEnvWrapper(
        ClassicAtariEnv('BreakoutNoFrameskip-v4'), frame_history=4
    ).instantiate(parallel_envs=16, seed=seed)

    # Again, use a helper to create a model
    # But because model is owned by the reinforcer, model should not be accessed using this variable
    # but from reinforcer.model property
    model_factory = QPolicyGradientModelFactory(
        backbone=NatureCnnFactory(input_width=84, input_height=84, input_channels=4)
    )

    # Reinforcer - an object managing the learning process
    reinforcer = BufferedMixedPolicyIterationReinforcer(
        device=device,
        settings=BufferedMixedPolicyIterationReinforcerSettings(
            discount_factor=0.99,
            experience_replay=2,
            stochastic_experience_replay=False
        ),
        model=model_factory.instantiate(action_space=vec_env.action_space),
        env=vec_env,
        algo=AcerPolicyGradient(
            model_factory=model_factory,
            entropy_coefficient=0.01,
            q_coefficient=0.5,
            rho_cap=10.0,
            retrace_rho_cap=1.0,
            trust_region=True,
            trust_region_delta=1.0,
            max_grad_norm=10.0,
        ),
        env_roller=ReplayQEnvRoller(
            environment=vec_env,
            device=device,
            number_of_steps=12,
            discount_factor=0.99,
            buffer_capacity=100,
            buffer_initial_size=100,
            frame_stack_compensation=4
        ),
    )

    # Model optimizer
    optimizer = optim.RMSprop(reinforcer.model.parameters(), lr=7.0e-4, eps=1e-3, alpha=0.99)

    # Overall information store for training information
    training_info = TrainingInfo(
        metrics=[
            EpisodeRewardMetric('episode_rewards'),  # Calculate average reward from episode
        ],
        callbacks=[]  # Print live metrics every epoch to standard output
    )

    # A bit of training initialization bookkeeping...
    training_info.initialize()
    reinforcer.initialize_training(training_info)
    training_info.on_train_begin()

    # Let's make 100 batches per epoch to average metrics nicely
    num_epochs = 1

    # Normal handrolled training loop
    for i in range(1, num_epochs+1):
        epoch_info = EpochInfo(
            training_info=training_info,
            global_epoch_idx=i,
            batches_per_epoch=1,
            optimizer=optimizer
        )

        reinforcer.train_epoch(epoch_info, interactive=False)

    training_info.on_train_end()





import attr
import sys
import tqdm

import gym
import torch

from vel.api import BatchInfo, EpochInfo
from vel.api.base import Model, ModelFactory
from vel.rl.api.base import ReinforcerBase, ReinforcerFactory, EnvFactory, ReplayEnvRollerBase, AlgoBase
from vel.rl.api.base.env_roller import ReplayEnvRollerFactory
from vel.rl.metrics import (
    FPSMetric, EpisodeLengthMetric, EpisodeRewardMetricQuantile, EpisodeRewardMetric, FramesMetric,
)


@attr.s(auto_attribs=True)
class BufferedSingleOffPolicyIterationReinforcerSettings:
    """ Settings class for deep Q-Learning """
    batch_rollout_rounds: int
    batch_training_rounds: int

    batch_size: int
    discount_factor: float


class BufferedSingleOffPolicyIterationReinforcer(ReinforcerBase):
    """
    An off-policy reinforcer that rolls out **single** environment and stores transitions in a buffer.
    Afterwards, it samples experience batches from this buffer to train the policy.
    """
    def __init__(self, device: torch.device, settings: BufferedSingleOffPolicyIterationReinforcerSettings,
                 environment: gym.Env, model: Model, algo: AlgoBase, env_roller: ReplayEnvRollerBase):
        self.device = device
        self.settings = settings
        self.environment = environment

        self._trained_model = model.to(self.device)
        self.algo = algo

        self.env_roller = env_roller

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        my_metrics = [
            FramesMetric("frames"),
            FPSMetric("fps"),
            EpisodeRewardMetric('PMM:episode_rewards'),
            EpisodeRewardMetricQuantile('P09:episode_rewards', quantile=0.9),
            EpisodeRewardMetricQuantile('P01:episode_rewards', quantile=0.1),
            EpisodeLengthMetric("episode_length")
        ]

        return my_metrics + self.algo.metrics() + self.env_roller.metrics()

    @property
    def model(self) -> Model:
        return self._trained_model

    def initialize_training(self, training_info):
        """ Prepare models for training """
        self.model.reset_weights()
        self.algo.initialize(self.settings, model=self.model, environment=self.environment, device=self.device)

    def train_epoch(self, epoch_info: EpochInfo, interactive=True) -> None:
        """ Train model for a single epoch  """
        epoch_info.on_epoch_begin()

        if interactive:
            iterator = tqdm.trange(epoch_info.batches_per_epoch, file=sys.stdout, desc="Training", unit="batch")
        else:
            iterator = range(epoch_info.batches_per_epoch)

        for batch_idx in iterator:
            batch_info = BatchInfo(epoch_info, batch_idx)

            batch_info.on_batch_begin()
            self.train_batch(batch_info)
            batch_info.on_batch_end()

        epoch_info.result_accumulator.freeze_results()
        epoch_info.on_epoch_end()

    def train_batch(self, batch_info: BatchInfo) -> None:
        """
        Batch - the most atomic unit of learning.

        For this reinforforcer, that involves:

        1. Roll out environment and store out experience in the buffer
        2. Sample the buffer and train the algo on sample batch
        """
        # Each DQN batch is
        # 1. Roll out environment and store out experience in the buffer
        self.model.eval()

        # Helper variables for rollouts
        episode_information = []
        frames = 0

        with torch.no_grad():
            if not self.env_roller.is_ready_for_sampling():
                while not self.env_roller.is_ready_for_sampling():
                    rollout = self.env_roller.rollout(batch_info, self.model)

                    episode_information.extend(rollout.episode_information())
                    frames += rollout.frames()
            else:
                for i in range(self.settings.batch_rollout_rounds):
                    rollout = self.env_roller.rollout(batch_info, self.model)

                    episode_information.extend(rollout.episode_information())
                    frames += rollout.frames()

        batch_info['frames'] = frames
        batch_info['episode_infos'] = episode_information

        # 2. Sample the buffer and train the algo on sample batch
        self.model.train()

        # Algo will aggregate data into this list:
        batch_info['sub_batch_data'] = []

        for i in range(self.settings.batch_training_rounds):
            sampled_rollout = self.env_roller.sample(batch_info, self.model)

            batch_result = self.algo.optimizer_step(
                batch_info=batch_info,
                device=self.device,
                model=self.model,
                rollout=sampled_rollout
            )

            self.env_roller.update(rollout=sampled_rollout, batch_info=batch_result)

            batch_info['sub_batch_data'].append(batch_result)

        batch_info.aggregate_key('sub_batch_data')


class BufferedSingleOffPolicyIterationReinforcerFactory(ReinforcerFactory):
    """ Factory class for the DQN reinforcer """

    def __init__(self, settings, env_factory: EnvFactory, model_factory: ModelFactory,
                 algo: AlgoBase, env_roller_factory: ReplayEnvRollerFactory, seed: int):
        self.settings = settings

        self.env_factory = env_factory
        self.model_factory = model_factory
        self.algo = algo
        self.env_roller_factory = env_roller_factory
        self.seed = seed

    def instantiate(self, device: torch.device) -> BufferedSingleOffPolicyIterationReinforcer:
        env = self.env_factory.instantiate(seed=self.seed)
        env_roller = self.env_roller_factory.instantiate(env, device, self.settings)
        model = self.model_factory.instantiate(action_space=env.action_space)

        return BufferedSingleOffPolicyIterationReinforcer(
            device=device,
            settings=self.settings,
            environment=env,
            model=model,
            algo=self.algo,
            env_roller=env_roller
        )


def create(model_config, env, model, algo, env_roller, batch_size: int, discount_factor: float,
           batch_rollout_rounds=1, batch_training_rounds=1):
    """ Vel creation function for DqnReinforcerFactory """
    settings = BufferedSingleOffPolicyIterationReinforcerSettings(
        batch_rollout_rounds=batch_rollout_rounds,
        batch_training_rounds=batch_training_rounds,
        batch_size=batch_size,
        discount_factor=discount_factor
    )

    return BufferedSingleOffPolicyIterationReinforcerFactory(
        settings=settings,
        env_factory=env,
        model_factory=model,
        algo=algo,
        env_roller_factory=env_roller,
        seed=model_config.seed
    )



import visdom
import pandas as pd


import vel.api.base as base

from vel.api import ModelConfig
from vel.util.visdom import visdom_append_metrics, VisdomSettings


class VisdomStreaming(base.Callback):
    """ Stream live results to visdom from training """
    def __init__(self, model_config: ModelConfig, visdom_settings: VisdomSettings):
        self.model_config = model_config
        self.settings = visdom_settings
        self.vis = visdom.Visdom(
            server=visdom_settings.server,
            endpoint=visdom_settings.endpoint,
            port=visdom_settings.port,
            env=self.model_config.run_name.replace('/', '_')
        )

    def on_epoch_end(self, epoch_info):
        """ Update data in visdom on push """
        metrics_df = pd.DataFrame([epoch_info.result]).set_index('epoch_idx')

        visdom_append_metrics(
            self.vis,
            metrics_df,
            first_epoch=epoch_info.global_epoch_idx == 1
        )

    def on_batch_end(self, batch_info):
        """ Stream LR to visdom """
        if self.settings.stream_lr:
            iteration_idx = (
                    float(batch_info.epoch_number) +
                    float(batch_info.batch_number) / batch_info.batches_per_epoch
            )
            
            lr = batch_info.optimizer.param_groups[-1]['lr']

            metrics_df = pd.DataFrame([lr], index=[iteration_idx], columns=['lr'])

            visdom_append_metrics(
                self.vis,
                metrics_df,
                first_epoch=(batch_info.epoch_number == 1) and (batch_info.batch_number == 0)
            )


def create(model_config, visdom_settings):
    return VisdomStreaming(model_config, VisdomSettings(**visdom_settings))



import numpy as np

import torch
import torch.nn as nn
import torch.nn.init as init
import torch.nn.functional as F

import gym.spaces as spaces


class DiagGaussianActionHead(nn.Module):
    """
    Action head where actions are normally distibuted uncorrelated variables with specific means and variances.

    Means are calculated directly from the network while standard deviation are a parameter of this module
    """

    LOG2PI = np.log(2.0 * np.pi)

    def __init__(self, input_dim, num_dimensions):
        super().__init__()

        self.input_dim = input_dim
        self.num_dimensions = num_dimensions

        self.linear_layer = nn.Linear(input_dim, num_dimensions)
        self.log_std = nn.Parameter(torch.zeros(1, num_dimensions))

    def forward(self, input_data):
        means = self.linear_layer(input_data)
        log_std_tile = self.log_std.repeat(means.size(0), 1)

        return torch.stack([means, log_std_tile], dim=-1)

    def sample(self, params, argmax_sampling=False):
        """ Sample from a probability space of all actions """
        means = params[:, :, 0]
        log_std = params[:, :, 1]

        if argmax_sampling:
            return means
        else:
            return torch.randn_like(means) * torch.exp(log_std) + means

    def logprob(self, action_sample, pd_params):
        """ Log-likelihood """
        means = pd_params[:, :, 0]
        log_std = pd_params[:, :, 1]

        std = torch.exp(log_std)

        z_score = (action_sample - means) / std

        return - (0.5 * ((z_score**2 + self.LOG2PI).sum(dim=-1)) + log_std.sum(dim=-1))

    def reset_weights(self):
        init.orthogonal_(self.linear_layer.weight, gain=0.01)
        init.constant_(self.linear_layer.bias, 0.0)

    def entropy(self, params):
        """
        Categorical distribution entropy calculation - sum probs * log(probs).
        In case of diagonal gaussian distribution - 1/2 log(2 pi e sigma^2)
        """
        log_std = params[:, :, 1]
        return (log_std + 0.5 * (self.LOG2PI + 1)).sum(dim=-1)

    def kl_divergence(self, params_q, params_p):
        """
        Categorical distribution KL divergence calculation
        KL(Q || P) = sum Q_i log (Q_i / P_i)

        Formula is:
        log(sigma_p) - log(sigma_q) + (sigma_q^2 + (mu_q - mu_p)^2))/(2 * sigma_p^2)
        """
        means_q = params_q[:, :, 0]
        log_std_q = params_q[:, :, 1]

        means_p = params_p[:, :, 0]
        log_std_p = params_p[:, :, 1]

        std_q = torch.exp(log_std_q)
        std_p = torch.exp(log_std_p)

        kl_div = log_std_p - log_std_q + (std_q ** 2 + (means_q - means_p) ** 2) / (2.0 * std_p ** 2) - 0.5

        return kl_div.sum(dim=-1)


class CategoricalActionHead(nn.Module):
    """ Action head with categorical actions """
    def __init__(self, input_dim, num_actions):
        super().__init__()

        self.input_dim = input_dim
        self.num_actions = num_actions

        self.linear_layer = nn.Linear(input_dim, num_actions)

    def forward(self, input_data):
        return F.log_softmax(self.linear_layer(input_data), dim=1)

    def logprob(self, actions, action_logits):
        """ Logarithm of probability of given sample """
        neg_log_prob = F.nll_loss(action_logits, actions, reduction='none')
        return -neg_log_prob

    def sample(self, logits, argmax_sampling=False):
        """ Sample from a probability space of all actions """
        if argmax_sampling:
            return torch.argmax(logits, dim=-1)
        else:
            u = torch.rand_like(logits)
            return torch.argmax(logits - torch.log(-torch.log(u)), dim=-1)

    def reset_weights(self):
        init.orthogonal_(self.linear_layer.weight, gain=0.01)
        init.constant_(self.linear_layer.bias, 0.0)

    def entropy(self, logits):
        """ Categorical distribution entropy calculation - sum probs * log(probs) """
        probs = torch.exp(logits)
        entropy = - torch.sum(probs * logits, dim=-1)
        return entropy

    def kl_divergence(self, logits_q, logits_p):
        """
        Categorical distribution KL divergence calculation
        KL(Q || P) = sum Q_i log (Q_i / P_i)
        When talking about logits this is:
        sum exp(Q_i) * (Q_i - P_i)
        """
        return (torch.exp(logits_q) * (logits_q - logits_p)).sum(1, keepdim=True)


class ActionHead(nn.Module):
    """
    Network head for action determination. Returns probability distribution parametrization
    """

    def __init__(self, input_dim, action_space):
        super().__init__()

        self.action_space = action_space

        if isinstance(action_space, spaces.Box):
            assert len(action_space.shape) == 1
            self.head = DiagGaussianActionHead(input_dim, action_space.shape[0])
        elif isinstance(action_space, spaces.Discrete):
            self.head = CategoricalActionHead(input_dim, action_space.n)
        # elif isinstance(action_space, spaces.MultiDiscrete):
        #     return MultiCategoricalPdType(action_space.nvec)
        # elif isinstance(action_space, spaces.MultiBinary):
        #     return BernoulliPdType(action_space.n)
        else:
            raise NotImplementedError

    def forward(self, input_data):
        return self.head(input_data)

    def sample(self, policy_params, **kwargs):
        """ Sample from a probability space of all actions """
        return self.head.sample(policy_params, **kwargs)

    def reset_weights(self):
        """ Initialize weights to sane defaults """
        self.head.reset_weights()

    def entropy(self, policy_params):
        """ Entropy calculation - sum probs * log(probs) """
        return self.head.entropy(policy_params)

    def kl_divergence(self, params_q, params_p):
        """ Kullback–Leibler divergence between two sets of parameters """
        return self.head.kl_divergence(params_q, params_p)

    def logprob(self, action_sample, policy_params):
        """ - log probabilty of selected actions """
        return self.head.logprob(action_sample, policy_params)



from vel.api.metrics.averaging_metric import AveragingSupervisedMetric


class Accuracy(AveragingSupervisedMetric):
    """ Classification accuracy """
    def __init__(self):
        super().__init__("accuracy")

    def _value_function(self, x_input, y_true, y_pred):
        """ Return classification accuracy of input """
        if len(y_true.shape) == 1:
            return y_pred.argmax(1).eq(y_true).double().mean().item()
        else:
            raise NotImplementedError


def create():
    """ Create accuracy metric """
    return Accuracy()



import torchvision.models.resnet as m
import torch.nn as nn
import torch.nn.functional as F

import vel.modules.layers as l
import vel.util.module_util as mu

from vel.api.base import SupervisedModel, ModelFactory


# Because of concat pooling it's 2x 512
NET_OUTPUT = 1024


class Resnet34(SupervisedModel):
    """ Resnet34 network model """

    def __init__(self, fc_layers=None, dropout=None, pretrained=True):
        super().__init__()

        # Store settings, maybe someone will be interested to see them
        self.fc_layers = fc_layers
        self.dropout = dropout
        self.pretrained = pretrained

        self.head_layers = 8
        self.group_cut_layers = (6, 10)

        # Load backbbone
        backbone = m.resnet34(pretrained=pretrained)

        # If fc layers is set, let's put custom head
        if fc_layers:
            # Take out the old head and let's put the new head
            valid_children = list(backbone.children())[:-2]

            valid_children.extend([
                l.AdaptiveConcatPool2d(),
                l.Flatten()
            ])

            layer_inputs = [NET_OUTPUT] + fc_layers[:-1]

            dropout = dropout or [None] * len(fc_layers)

            for idx, (layer_input, layet_output, layer_dropout) in enumerate(zip(layer_inputs, fc_layers, dropout)):
                valid_children.append(nn.BatchNorm1d(layer_input))

                if layer_dropout:
                    valid_children.append(nn.Dropout(layer_dropout))

                valid_children.append(nn.Linear(layer_input, layet_output))

                if idx == len(fc_layers) - 1:
                    # Last layer
                    valid_children.append(nn.LogSoftmax(dim=1))
                else:
                    valid_children.append(nn.ReLU())

            final_model = nn.Sequential(*valid_children)
        else:
            final_model = backbone

        self.model = final_model

    def freeze(self, number=None):
        """ Freeze given number of layers in the model """
        if number is None:
            number = self.head_layers

        for idx, child in enumerate(self.model.children()):
            if idx < number:
                mu.freeze_layer(child)

    def unfreeze(self):
        """ Unfreeze model layers """
        for idx, child in enumerate(self.model.children()):
            mu.unfreeze_layer(child)

    def get_layer_groups(self):
        """ Return layers grouped """
        g1 = list(self.model[:self.group_cut_layers[0]])
        g2 = list(self.model[self.group_cut_layers[0]:self.group_cut_layers[1]])
        g3 = list(self.model[self.group_cut_layers[1]:])
        return [g1, g2, g3]

    def forward(self, x):
        """ Calculate model value """
        return self.model(x)

    def loss_value(self, x_data, y_true, y_pred):
        """ Calculate value of the loss function """
        return F.nll_loss(y_pred, y_true)

    def metrics(self):
        """ Set of metrics for this model """
        from vel.metrics.loss_metric import Loss
        from vel.metrics.accuracy import Accuracy
        return [Loss(), Accuracy()]


def create(fc_layers=None, dropout=None, pretrained=True):
    """ Create a Resnet-34 model with a custom head """
    def instantiate(**_):
        return Resnet34(fc_layers, dropout, pretrained)

    return ModelFactory.generic(instantiate)



import numpy as np
from gym import spaces
from collections import OrderedDict
from . import VecEnv

class DummyVecEnv(VecEnv):
    def __init__(self, env_fns):
        self.envs = [fn() for fn in env_fns]
        env = self.envs[0]
        VecEnv.__init__(self, len(env_fns), env.observation_space, env.action_space)
        shapes, dtypes = {}, {}
        self.keys = []
        obs_space = env.observation_space

        if isinstance(obs_space, spaces.Dict):
            assert isinstance(obs_space.spaces, OrderedDict)
            subspaces = obs_space.spaces
        else:
            subspaces = {None: obs_space}

        for key, box in subspaces.items():
            shapes[key] = box.shape
            dtypes[key] = box.dtype
            self.keys.append(key)

        self.buf_obs = { k: np.zeros((self.num_envs,) + tuple(shapes[k]), dtype=dtypes[k]) for k in self.keys }
        self.buf_dones = np.zeros((self.num_envs,), dtype=np.bool)
        self.buf_rews  = np.zeros((self.num_envs,), dtype=np.float32)
        self.buf_infos = [{} for _ in range(self.num_envs)]
        self.actions = None

    def step_async(self, actions):
        self.actions = actions

    def step_wait(self):
        for e in range(self.num_envs):
            obs, self.buf_rews[e], self.buf_dones[e], self.buf_infos[e] = self.envs[e].step(self.actions[e])
            if self.buf_dones[e]:
                obs = self.envs[e].reset()
            self._save_obs(e, obs)
        return (self._obs_from_buf(), np.copy(self.buf_rews), np.copy(self.buf_dones),
                self.buf_infos.copy())

    def reset(self):
        for e in range(self.num_envs):
            obs = self.envs[e].reset()
            self._save_obs(e, obs)
        return self._obs_from_buf()

    def close(self):
        return

    def render(self, mode='human'):
        return [e.render(mode=mode) for e in self.envs]

    def _save_obs(self, e, obs):
        for k in self.keys:
            if k is None:
                self.buf_obs[k][e] = obs
            else:
                self.buf_obs[k][e] = obs[k]

    def _obs_from_buf(self):
        if self.keys==[None]:
            return self.buf_obs[None]
        else:
            return self.buf_obs



import pymongo
import pandas as pd


class MongoDbBackend:
    """ Storage backend to store all experiment results in a MongoDB database """

    def __init__(self, model_config, uri, database):
        self.model_config = model_config
        self.client = pymongo.MongoClient(uri)
        self.db = self.client[database]

    def clean(self, initial_epoch):
        """ Remove entries from database that would get overwritten """
        self.db.metrics.delete_many({'run_name': self.model_config.run_name, 'epoch_idx': {'$gt': initial_epoch}})

    def store_config(self, configuration):
        """ Store model parameters in the database """
        run_name = self.model_config.run_name

        self.db.configs.delete_many({'run_name': self.model_config.run_name})

        configuration = configuration.copy()
        configuration['run_name'] = run_name

        self.db.configs.insert_one(configuration)

    def get_frame(self):
        """ Get a dataframe of metrics from this storage """
        metric_items = list(self.db.metrics.find({'run_name': self.model_config.run_name}).sort('epoch_idx'))
        if len(metric_items) == 0:
            return pd.DataFrame(columns=['run_name'])
        else:
            return pd.DataFrame(metric_items).drop(['_id', 'model_name'], axis=1).set_index('epoch_idx')

    def store(self, metrics):
        augmented_metrics = metrics.copy()

        model_name = self.model_config.name
        run_name = self.model_config.run_name

        augmented_metrics['model_name'] = model_name
        augmented_metrics['run_name'] = run_name

        self.db.metrics.insert_one(augmented_metrics)


def create(model_config, uri, database):
    """ Create a mongodb storage object """
    return MongoDbBackend(model_config, uri, database)




from ..info import EpochInfo, BatchInfo, TrainingInfo


class Callback:
    """
    An abstract class that all callback classes extends from.
    """

    def on_initialization(self, training_info: TrainingInfo) -> None:
        """
        Runs for the first time a training process is started from scratch. Is guaranteed to be run only once
        for the training process. Will be run before all other callbacks.
        """
        pass

    def on_train_begin(self, training_info: TrainingInfo) -> None:
        """
        Beginning of a training process - is run every time a training process is started, even if it's restarted from
        a checkpoint.
        """
        pass

    def on_train_end(self, training_info: TrainingInfo) -> None:
        """
        Finalize training process. Runs each time at the end of a training process.
        """
        pass

    def on_epoch_begin(self, epoch_info: EpochInfo) -> None:
        """
        Run for each epoch before an epoch will be trained
        """
        pass

    def on_epoch_end(self, epoch_info: EpochInfo) -> None:
        """
        Runs for each epoch after an epoch is trained
        """
        pass

    def on_batch_begin(self, batch_info: BatchInfo) -> None:
        """
        Runs for each batch before batch is evaluated
        """
        pass

    def on_batch_end(self, batch_info: BatchInfo) -> None:
        """
        Runs for each batch after batch is evaluated
        """
        pass

    def on_validation_batch_begin(self, batch_info: BatchInfo) -> None:
        """
        Supervised learning only - runs before validation batch
        """
        pass

    def on_validation_batch_end(self, batch_info: BatchInfo) -> None:
        """
        Supervised learning only - runs after validation batch
        """
        pass

    def write_state_dict(self, training_info: TrainingInfo, hidden_state_dict: dict) -> None:
        """
        Persist callback state to the state dictionary
        """
        pass

    def load_state_dict(self, training_info: TrainingInfo, hidden_state_dict: dict) -> None:
        """
        Load callback state from the state dictionary
        """
        pass

    # TODO(jerry): provide some info about current phase
    # def on_phase_begin(self) -> None: pass
    #
    # def on_phase_end(self) -> None: pass



from vel.api import EpochInfo, TrainingInfo
from vel.api.base import Model


class Storage:
    """ Base class for Vel storage implementations """

    def last_epoch_idx(self) -> int:
        """ Return last checkpointed epoch idx for given configuration. Returns 0 if no results have been stored """
        raise NotImplementedError

    def resume(self, train_info: TrainingInfo, model: Model) -> dict:
        """
        Resume learning process and return loaded hidden state dictionary
        """
        raise NotImplementedError

    def reset(self, configuration: dict) -> None:
        """
        Whenever there was anything stored in the database or not, purge previous state and start
        new training process from scratch.
        """
        raise NotImplementedError

    def streaming_callbacks(self) -> list:
        """ Lift of callbacks for live streaming results """
        return []

    def get_metrics_frame(self):
        """ Get a frame of metrics from backend """
        raise NotImplementedError

    def checkpoint(self, epoch_info: EpochInfo, model: Model):
        """ When epoch is done, we persist the training state """
        raise NotImplementedError



from vel.api import ModelConfig


def load(config_path, run_number=0, device='cuda:0'):
    """ Load a ModelConfig from filename """
    model_config = ModelConfig.from_file(config_path, run_number, device=device)

    return model_config



import torch
import torch.nn.functional as F
import torch.nn.utils

from vel.api.base import ModelFactory
from vel.api.metrics.averaging_metric import AveragingNamedMetric
from vel.rl.api.base import OptimizerAlgoBase


class DeepQLearning(OptimizerAlgoBase):
    """ Deep Q-Learning algorithm """

    def __init__(self, model_factory: ModelFactory, double_dqn: bool,
                 target_update_frequency: int, max_grad_norm: float):
        super().__init__(max_grad_norm)

        self.model_factory = model_factory

        self.double_dqn = double_dqn
        self.target_update_frequency = target_update_frequency

        self.discount_factor = None
        self.target_model = None

    def initialize(self, settings, model, environment, device):
        """ Initialize policy gradient from reinforcer settings """
        self.target_model = self.model_factory.instantiate(action_space=environment.action_space).to(device)
        self.target_model.load_state_dict(model.state_dict())
        self.target_model.eval()

        self.discount_factor = settings.discount_factor

    def calculate_gradient(self, batch_info, device, model, rollout):
        """ Calculate loss of the supplied rollout """
        evaluator = model.evaluate(rollout)

        dones_tensor = evaluator.get('rollout:dones')
        rewards_tensor = evaluator.get('rollout:rewards')

        assert dones_tensor.dtype == torch.float32

        with torch.no_grad():
            target_evaluator = self.target_model.evaluate(rollout)

            if self.double_dqn:
                # DOUBLE DQN
                target_q = target_evaluator.get('model:q_next')
                model_q = evaluator.get('model:q_next')
                # Select largest 'target' value based on action that 'model' selects
                values = target_q.gather(1, model_q.argmax(dim=1, keepdim=True)).squeeze(1)
            else:
                # REGULAR DQN
                # [0] is because in pytorch .max(...) returns tuple (max values, argmax)
                values = target_evaluator.get('model:q_next').max(dim=1)[0]

            estimated_return = rewards_tensor + self.discount_factor * values * (1 - dones_tensor)

        q_selected = evaluator.get('model:action:q')
        weights = evaluator.get('rollout:weights')

        original_losses = F.smooth_l1_loss(q_selected, estimated_return, reduction='none')

        loss_value = torch.mean(weights * original_losses)
        loss_value.backward()

        return {
            'loss': loss_value.item(),
            # We need it to update priorities in the replay buffer:
            'errors': original_losses.detach().cpu().numpy(),
            'average_q_selected': torch.mean(q_selected).item(),
            'average_q_target': torch.mean(estimated_return).item()
        }

    def post_optimization_step(self, batch_info, device, model, rollout):
        """ Steps to take after optimization has been done"""
        if batch_info.aggregate_batch_number % self.target_update_frequency == 0:
            self.target_model.load_state_dict(model.state_dict())
            self.target_model.eval()

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        return [
            AveragingNamedMetric("loss"),
            AveragingNamedMetric("average_q_selected"),
            AveragingNamedMetric("average_q_target"),
            AveragingNamedMetric("grad_norm"),
        ]


def create(model: ModelFactory, target_update_frequency: int,
           max_grad_norm: float, double_dqn: bool=False):
    return DeepQLearning(
        model_factory=model,
        double_dqn=double_dqn,
        target_update_frequency=target_update_frequency,
        max_grad_norm=max_grad_norm
    )



import hashlib
import torch.nn as nn

import vel.util.module_util as mu

from vel.metrics.loss_metric import Loss
from vel.util.summary import summary


class Model(nn.Module):
    """ Class representing full neural network model """

    def metrics(self) -> list:
        """ Set of metrics for this model """
        return [Loss()]

    def train(self, mode=True):
        r"""
        Sets the module in training mode.

        This has any effect only on certain modules. See documentations of
        particular modules for details of their behaviors in training/evaluation
        mode, if they are affected, e.g. :class:`Dropout`, :class:`BatchNorm`,
        etc.

        Returns:
            Module: self
        """
        super().train(mode)

        if mode:
            mu.apply_leaf(self, mu.set_train_mode)

        return self

    def summary(self, input_size=None, hashsummary=False):
        """ Print a model summary """

        if input_size is None:
            print(self)
            print("-" * 120)
            number = sum(p.numel() for p in self.model.parameters())
            print("Number of model parameters: {:,}".format(number))
            print("-" * 120)
        else:
            summary(self, input_size)

        if hashsummary:
            for idx, hashvalue in enumerate(self.hashsummary()):
                print(f"{idx}: {hashvalue}")

    def hashsummary(self):
        """ Print a model summary - checksums of each layer parameters """
        children = list(self.children())

        result = []

        for child in children:
            result.extend(hashlib.sha256(x.detach().cpu().numpy().tobytes()).hexdigest() for x in child.parameters())

        return result

    def get_layer_groups(self):
        """ Return layers grouped """
        return [self]

    def reset_weights(self):
        """ Call proper initializers for the weights """
        pass

    @property
    def is_recurrent(self) -> bool:
        """ If the network is recurrent and needs to be fed state as well as the observations """
        return False


class RnnModel(Model):
    """ Class representing recurrent model """

    @property
    def is_recurrent(self) -> bool:
        """ If the network is recurrent and needs to be fed previous state """
        return True

    @property
    def state_dim(self) -> int:
        """ Dimension of model state """
        raise NotImplementedError


class BackboneModel(Model):
    """ Model that serves as a backbone network to connect your heads to """


class RnnLinearBackboneModel(BackboneModel):
    """
    Model that serves as a backbone network to connect your heads to -
    one that spits out a single-dimension output and is a recurrent neural network
    """

    @property
    def is_recurrent(self) -> bool:
        """ If the network is recurrent and needs to be fed previous state """
        return True

    @property
    def output_dim(self) -> int:
        """ Final dimension of model output """
        raise NotImplementedError

    @property
    def state_dim(self) -> int:
        """ Dimension of model state """
        raise NotImplementedError


class LinearBackboneModel(BackboneModel):
    """
    Model that serves as a backbone network to connect your heads to - one that spits out a single-dimension output
    """

    @property
    def output_dim(self) -> int:
        """ Final dimension of model output """
        raise NotImplementedError


class SupervisedModel(Model):
    """ Model for a supervised learning probvlem"""
    def loss(self, x_data, y_true):
        """ Forward propagate network and return a value of loss function """
        y_pred = self(x_data)
        return y_pred, self.loss_value(x_data, y_true, y_pred)

    def loss_value(self, x_data, y_true, y_pred):
        """ Calculate a value of loss function """
        raise NotImplementedError



from .base_metric import BaseMetric


class SummingMetric(BaseMetric):
    """ Base class for metrics that simply calculate the sum over the epoch """
    def __init__(self, name, reset_value=True):
        super().__init__(name)

        self.reset_value = reset_value
        self.buffer = 0

    def calculate(self, batch_info):
        """ Calculate value of a metric """
        value = self._value_function(batch_info)
        self.buffer += value

    def _value_function(self, batch_info):
        raise NotImplementedError

    def reset(self):
        """ Reset value of a metric """

        if self.reset_value:
            self.buffer = 0

    def value(self):
        """ Return current value for the metric """
        return self.buffer


class SummingNamedMetric(SummingMetric):
    """ Super simple summing metric that just takes a value from dictionary and sums it over samples """
    def __init__(self, name, reset_value=True):
        super().__init__(name, reset_value=reset_value)

    def _value_function(self, batch_info):
        return batch_info[self.name].item()



from vel.api.base import RnnLinearBackboneModel, ModelFactory


class LstmBackbone(RnnLinearBackboneModel):
    """
    Simple 'LSTM' model backbone
    """

    def __init__(self, input_size, hidden_units):
        super().__init__()

        self.input_size = input_size
        self.hidden_units = hidden_units

    def forward(self, input_data, masks, state):
        pass

    def initial_state(self):
        """ Initial state of the network """
        pass



import attr
import numpy as np
import sys
import torch
import tqdm

from vel.api.base import Model, ModelFactory
from vel.api.info import EpochInfo, BatchInfo
from vel.rl.api.base import ReinforcerBase, ReinforcerFactory, VecEnvFactory, EnvRollerFactory, EnvRollerBase, AlgoBase
from vel.rl.metrics import (
    FPSMetric, EpisodeLengthMetric, EpisodeRewardMetricQuantile,
    EpisodeRewardMetric, FramesMetric
)


@attr.s(auto_attribs=True)
class OnPolicyIterationReinforcerSettings:
    """ Settings dataclass for a policy gradient reinforcer """
    discount_factor: float

    batch_size: int = 256
    experience_replay: int = 1
    stochastic_experience_replay: bool = False

    # For each experience replay loop, shuffle transitions to randomize gradient calculations
    # That means, disregarding actual trajectory order
    # Does not work with RNN policies
    shuffle_transitions: bool = True


class OnPolicyIterationReinforcer(ReinforcerBase):
    """
    A reinforcer that calculates on-policy environment rollouts and uses them to train policy directly.
    May split the sample into multiple batches and may replay batches a few times.
    """
    def __init__(self, device: torch.device, settings: OnPolicyIterationReinforcerSettings, model: Model,
                 algo: AlgoBase, env_roller: EnvRollerBase) -> None:
        self.device = device
        self.settings = settings

        self._trained_model = model.to(self.device)

        self.env_roller = env_roller
        self.algo = algo

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        my_metrics = [
            FramesMetric("frames"),
            FPSMetric("fps"),
            EpisodeRewardMetric('PMM:episode_rewards'),
            EpisodeRewardMetricQuantile('P09:episode_rewards', quantile=0.9),
            EpisodeRewardMetricQuantile('P01:episode_rewards', quantile=0.1),
            EpisodeLengthMetric("episode_length"),
        ]

        return my_metrics + self.algo.metrics() + self.env_roller.metrics()

    @property
    def model(self) -> Model:
        """ Model trained by this reinforcer """
        return self._trained_model

    def initialize_training(self, training_info):
        """ Prepare models for training """
        self.model.reset_weights()
        self.algo.initialize(
            self.settings, model=self.model, environment=self.env_roller.environment, device=self.device
        )

    def train_epoch(self, epoch_info: EpochInfo, interactive=True) -> None:
        """ Train model on an epoch of a fixed number of batch updates """
        epoch_info.on_epoch_begin()

        if interactive:
            iterator = tqdm.trange(epoch_info.batches_per_epoch, file=sys.stdout, desc="Training", unit="batch")
        else:
            iterator = range(epoch_info.batches_per_epoch)

        for batch_idx in iterator:
            batch_info = BatchInfo(epoch_info, batch_idx)

            batch_info.on_batch_begin()
            self.train_batch(batch_info)
            batch_info.on_batch_end()

        epoch_info.result_accumulator.freeze_results()
        epoch_info.on_epoch_end()

    def train_batch(self, batch_info: BatchInfo) -> None:
        """
        Batch - the most atomic unit of learning.

        For this reinforforcer, that involves:

        1. Roll out the environmnent using current policy
        2. Use that rollout to train the policy
        """
        # Calculate environment rollout on the evaluation version of the model
        self.model.eval()

        rollout = self.env_roller.rollout(batch_info, self.model)

        # Perform the training step
        self.model.train()

        # Algo will aggregate data into this list:
        batch_info['sub_batch_data'] = []

        if self.settings.shuffle_transitions:
            rollout = rollout.to_transitions()

        if self.settings.stochastic_experience_replay:
            # Always play experience at least once
            experience_replay_count = max(np.random.poisson(self.settings.experience_replay), 1)
        else:
            experience_replay_count = self.settings.experience_replay

        # Repeat the experience N times
        for i in range(experience_replay_count):
            # We may potentially need to split rollout into multiple batches
            if self.settings.batch_size >= rollout.frames():
                batch_result = self.algo.optimizer_step(
                    batch_info=batch_info,
                    device=self.device,
                    model=self.model,
                    rollout=rollout
                )

                batch_info['sub_batch_data'].append(batch_result)
            else:
                # Rollout too big, need to split in batches
                for batch_rollout in rollout.shuffled_batches(self.settings.batch_size):
                    batch_result = self.algo.optimizer_step(
                        batch_info=batch_info,
                        device=self.device,
                        model=self.model,
                        rollout=batch_rollout
                    )

                    batch_info['sub_batch_data'].append(batch_result)

        batch_info['frames'] = rollout.frames()
        batch_info['episode_infos'] = rollout.episode_information()

        # Even with all the experience replay, we count the single rollout as a single batch
        batch_info.aggregate_key('sub_batch_data')


class OnPolicyIterationReinforcerFactory(ReinforcerFactory):
    """ Vel factory class for the PolicyGradientReinforcer """
    def __init__(self, settings, parallel_envs: int, env_factory: VecEnvFactory, model_factory: ModelFactory,
                 algo: AlgoBase, env_roller_factory: EnvRollerFactory, seed: int):
        self.settings = settings
        self.parallel_envs = parallel_envs

        self.env_factory = env_factory
        self.model_factory = model_factory
        self.algo = algo
        self.env_roller_factory = env_roller_factory
        self.seed = seed

    def instantiate(self, device: torch.device) -> ReinforcerBase:
        env = self.env_factory.instantiate(parallel_envs=self.parallel_envs, seed=self.seed)
        model = self.model_factory.instantiate(action_space=env.action_space)
        env_roller = self.env_roller_factory.instantiate(environment=env, device=device, settings=self.settings)

        return OnPolicyIterationReinforcer(device, self.settings, model, self.algo, env_roller)


def create(model_config, model, vec_env, algo, env_roller, parallel_envs, discount_factor,
           batch_size=256, experience_replay=1, stochastic_experience_replay=False, shuffle_transitions=True):
    """ Create a policy gradient reinforcer - factory """
    settings = OnPolicyIterationReinforcerSettings(
        discount_factor=discount_factor,
        batch_size=batch_size,
        experience_replay=experience_replay,
        stochastic_experience_replay=stochastic_experience_replay,
        shuffle_transitions=shuffle_transitions
    )

    return OnPolicyIterationReinforcerFactory(
        settings=settings,
        parallel_envs=parallel_envs,
        env_factory=vec_env,
        model_factory=model,
        algo=algo,
        env_roller_factory=env_roller,
        seed=model_config.seed
    )



import typing

import torch
import torch.nn.functional as F
import torch.nn as nn

from vel.api.base import SupervisedModel, ModelFactory
from vel.modules.layers import OneHotEncode


class MultilayerSequenceLSTM(SupervisedModel):
    """ Multilayer LSTM network for sequence modeling """

    def __init__(self, alphabet_size: int, hidden_layers: typing.List[int], output_dim: int, use_embedding=False,
                 dropout: float=0.0):
        super().__init__()

        self.alphabet_size = alphabet_size
        self.output_dim = output_dim
        self.hidden_layers = hidden_layers

        input_dim = hidden_layers[0] if hidden_layers else output_dim

        if use_embedding:
            self.input_layer = nn.Embedding(alphabet_size, input_dim)
            current_dim = input_dim
        else:
            self.input_layer = nn.Sequential(
                OneHotEncode(alphabet_size),
                nn.Linear(alphabet_size, input_dim)
            )
            current_dim = input_dim

        self.lstm_layers = []
        self.dropout_layers = []

        for idx, current_layer in enumerate(hidden_layers, 1):
            lstm = nn.LSTM(
                input_size=current_dim,
                hidden_size=current_layer,
                batch_first=True,
                # dropout=dropout
            )

            self.add_module('lstm{:02}'.format(idx), lstm)
            self.lstm_layers.append(lstm)

            if dropout > 0.0:
                dropout_layer = nn.Dropout(p=dropout)

                self.add_module('dropout{:02}'.format(idx), dropout_layer)
                self.dropout_layers.append(dropout_layer)

            current_dim = current_layer

        self.output_layer = nn.Linear(current_dim, output_dim)
        self.output_activation = nn.LogSoftmax(dim=2)

    def forward(self, sequence):
        """ Forward propagate batch of sequences through the network, without accounting for the state """
        data = self.input_layer(sequence)

        for idx in range(len(self.lstm_layers)):
            data, _ = self.lstm_layers[idx](data)

            if self.dropout_layers:
                data = self.dropout_layers[idx](data)

        data = self.output_layer(data)

        return self.output_activation(data)

    def forward_state(self, sequence, state=None):
        """ Forward propagate a sequence through the network accounting for the state """
        if state is None:
            state = self.initial_state(sequence.size(0))

        data = self.input_layer(sequence)

        state_outputs = []

        for idx in range(len(self.lstm_layers)):
            layer_length = self.hidden_layers[idx]

            # Partition hidden state, for each layer we have layer_length of h state and layer_length of c state
            current_state = state[:, :, :layer_length * 2]
            state = state[:, :, 2 * layer_length:]

            # Split into h and c state
            current_h = current_state[:, :, :layer_length]
            current_c = current_state[:, :, layer_length:]

            # Propagate through the LSTM state
            data, (new_h, new_c) = self.lstm_layers[idx](data, (current_h, current_c))

            if self.dropout_layers:
                data = self.dropout_layers[idx](data)

            state_outputs.append(new_h)
            state_outputs.append(new_c)

        output_data = self.output_activation(self.output_layer(data))

        concatenated_hidden_output = torch.cat(state_outputs, dim=2)

        return output_data, concatenated_hidden_output

    def initial_state(self, batch_size):
        """ Initial state of the network """
        return torch.zeros(batch_size, 1, 2 * sum(self.hidden_layers))

    def loss_value(self, x_data, y_true, y_pred):
        """ Calculate a value of loss function """
        y_pred = y_pred.view(-1, y_pred.size(2))
        y_true = y_true.view(-1).to(torch.long)
        return F.nll_loss(y_pred, y_true)


def create(alphabet_size: int, hidden_layers: typing.List[int], output_dim: int, use_embedding=False, dropout=0.0):
    """ Vel creation function """
    def instantiate(**_):
        return MultilayerSequenceLSTM(
            alphabet_size, hidden_layers, output_dim, use_embedding=use_embedding, dropout=dropout
        )

    return ModelFactory.generic(instantiate)



"""
Code based on: https://github.com/sksq96/pytorch-summary/blob/master/torchsummary/torchsummary.py
"""
import torch
import torch.nn as nn
from torch.autograd import Variable

from collections import OrderedDict


def summary(model, input_size):
    """ Print summary of the model """
    def register_hook(module):
        def hook(module, input, output):
            class_name = str(module.__class__).split('.')[-1].split("'")[0]
            module_idx = len(summary)

            m_key = '%s-%i' % (class_name, module_idx + 1)
            summary[m_key] = OrderedDict()
            summary[m_key]['input_shape'] = list(input[0].size())
            summary[m_key]['input_shape'][0] = -1
            if isinstance(output, (list, tuple)):
                summary[m_key]['output_shape'] = [[-1] + list(o.size())[1:] for o in output]
            else:
                summary[m_key]['output_shape'] = list(output.size())
                summary[m_key]['output_shape'][0] = -1

            params = 0
            if hasattr(module, 'weight') and hasattr(module.weight, 'size'):
                params += torch.prod(torch.LongTensor(list(module.weight.size())))
                summary[m_key]['trainable'] = module.weight.requires_grad
            if hasattr(module, 'bias') and hasattr(module.bias, 'size'):
                params += torch.prod(torch.LongTensor(list(module.bias.size())))
            summary[m_key]['nb_params'] = params

        if (not isinstance(module, nn.Sequential) and
                not isinstance(module, nn.ModuleList) and
                not (module == model)):
            hooks.append(module.register_forward_hook(hook))

    if torch.cuda.is_available():
        dtype = torch.cuda.FloatTensor
        model = model.cuda()
    else:
        dtype = torch.FloatTensor
        model = model.cpu()

    # check if there are multiple inputs to the network
    if isinstance(input_size[0], (list, tuple)):
        x = [Variable(torch.rand(2, *in_size)).type(dtype) for in_size in input_size]
    else:
        x = Variable(torch.rand(2, *input_size)).type(dtype)

    # print(type(x[0]))
    # create properties
    summary = OrderedDict()
    hooks = []
    # register hook
    model.apply(register_hook)
    # make a forward pass
    # print(x.shape)
    model(x)
    # remove these hooks
    for h in hooks:
        h.remove()

    print('----------------------------------------------------------------')
    line_new = '{:>20}  {:>25} {:>15}'.format('Layer (type)', 'Output Shape', 'Param #')
    print(line_new)
    print('================================================================')
    total_params = 0
    trainable_params = 0
    for layer in summary:
        # input_shape, output_shape, trainable, nb_params
        line_new = '{:>20}  {:>25} {:>15}'.format(layer, str(summary[layer]['output_shape']),
                                                  '{0:,}'.format(summary[layer]['nb_params']))
        total_params += summary[layer]['nb_params']
        if 'trainable' in summary[layer]:
            if summary[layer]['trainable'] == True:
                trainable_params += summary[layer]['nb_params']
        print(line_new)
    print('================================================================')
    print('Total params: {0:,}'.format(total_params))
    print('Trainable params: {0:,}'.format(trainable_params))
    print('Non-trainable params: {0:,}'.format(total_params - trainable_params))
    print('----------------------------------------------------------------')
    # return summary


import gym
import numpy as np

from vel.openai.baselines.common.running_mean_std import RunningMeanStd


class EnvNormalize(gym.Wrapper):
    """
    Single environment normalization based on VecNormalize from OpenAI baselines
    """
    def __init__(self, env, normalize_observations=True, normalize_returns=True,
                 clip_observations=10., clip_rewards=10., gamma=0.99, epsilon=1e-8):
        super().__init__(env)

        self.ob_rms = RunningMeanStd(shape=self.observation_space.shape) if normalize_observations else None
        self.ret_rms = RunningMeanStd(shape=()) if normalize_returns else None
        self.clipob = clip_observations
        self.cliprew = clip_rewards
        self.ret = 0.0
        self.gamma = gamma
        self.epsilon = epsilon

    def step(self, action):
        """
        Apply sequence of actions to sequence of environments
        actions -> (observations, rewards, news)

        where 'news' is a boolean vector indicating whether each element is new.
        """
        obs, rews, news, infos = self.env.step(action)

        self.ret = self.ret * self.gamma + rews

        obs = self._filter_observation(obs)

        if self.ret_rms:
            self.ret_rms.update(np.array([self.ret]))
            rews = np.clip(rews / np.sqrt(self.ret_rms.var + self.epsilon), -self.cliprew, self.cliprew)

        return obs, rews, news, infos

    def _filter_observation(self, obs):
        if self.ob_rms:
            self.ob_rms.update(obs[None])
            obs = np.clip((obs - self.ob_rms.mean) / np.sqrt(self.ob_rms.var + self.epsilon), -self.clipob, self.clipob)

            return obs.astype(np.float32)
        else:
            return obs

    def reset(self):
        """
        Reset all environments
        """
        obs = self.env.reset()
        return self._filter_observation(obs)



__all__ = ['Monitor', 'get_monitor_files', 'load_results']

import gym
from gym.core import Wrapper
import time
from glob import glob
import csv
import os
import os.path as osp
import json
import uuid
import numpy as np
import pandas

class Monitor(Wrapper):
    EXT = "monitor.csv"
    f = None

    def __init__(self, env, filename, allow_early_resets=False, reset_keywords=(), info_keywords=()):
        Wrapper.__init__(self, env=env)
        self.tstart = time.time()
        if filename is None:
            self.f = None
            self.logger = None
        else:
            if not filename.endswith(Monitor.EXT):
                if osp.isdir(filename):
                    filename = osp.join(filename, Monitor.EXT)
                else:
                    filename = filename + "." + Monitor.EXT
            self.f = open(filename, "wt")
            self.f.write('#%s\n'%json.dumps({"t_start": self.tstart, 'env_id' : env.spec and env.spec.id}))
            self.logger = csv.DictWriter(self.f, fieldnames=('r', 'l', 't')+reset_keywords+info_keywords)
            self.logger.writeheader()
            self.f.flush()

        self.reset_keywords = reset_keywords
        self.info_keywords = info_keywords
        self.allow_early_resets = allow_early_resets
        self.rewards = None
        self.needs_reset = True
        self.episode_rewards = []
        self.episode_lengths = []
        self.episode_times = []
        self.total_steps = 0
        self.current_reset_info = {} # extra info about the current episode, that was passed in during reset()

    def reset(self, **kwargs):
        if not self.allow_early_resets and not self.needs_reset:
            raise RuntimeError("Tried to reset an environment before done. If you want to allow early resets, wrap your env with Monitor(env, path, allow_early_resets=True)")
        self.rewards = []
        self.needs_reset = False
        for k in self.reset_keywords:
            v = kwargs.get(k)
            if v is None:
                raise ValueError('Expected you to pass kwarg %s into reset'%k)
            self.current_reset_info[k] = v
        return self.env.reset(**kwargs)

    def step(self, action):
        if self.needs_reset:
            raise RuntimeError("Tried to step environment that needs reset")
        ob, rew, done, info = self.env.step(action)
        self.rewards.append(rew)
        if done:
            self.needs_reset = True
            eprew = sum(self.rewards)
            eplen = len(self.rewards)
            epinfo = {"r": round(eprew, 6), "l": eplen, "t": round(time.time() - self.tstart, 6)}
            for k in self.info_keywords:
                epinfo[k] = info[k]
            self.episode_rewards.append(eprew)
            self.episode_lengths.append(eplen)
            self.episode_times.append(time.time() - self.tstart)
            epinfo.update(self.current_reset_info)
            if self.logger:
                self.logger.writerow(epinfo)
                self.f.flush()
            info['episode'] = epinfo
        self.total_steps += 1
        return (ob, rew, done, info)

    def close(self):
        if self.f is not None:
            self.f.close()

    def get_total_steps(self):
        return self.total_steps

    def get_episode_rewards(self):
        return self.episode_rewards

    def get_episode_lengths(self):
        return self.episode_lengths

    def get_episode_times(self):
        return self.episode_times

class LoadMonitorResultsError(Exception):
    pass

def get_monitor_files(dir):
    return glob(osp.join(dir, "*" + Monitor.EXT))

def load_results(dir):
    import pandas
    monitor_files = (
            glob(osp.join(dir, "*monitor.json")) +
            glob(osp.join(dir, "*monitor.csv"))) # get both csv and (old) json files
    if not monitor_files:
        raise LoadMonitorResultsError("no monitor files of the form *%s found in %s" % (Monitor.EXT, dir))
    dfs = []
    headers = []
    for fname in monitor_files:
        with open(fname, 'rt') as fh:
            if fname.endswith('csv'):
                firstline = fh.readline()
                assert firstline[0] == '#'
                header = json.loads(firstline[1:])
                df = pandas.read_csv(fh, index_col=None)
                headers.append(header)
            elif fname.endswith('json'): # Deprecated json format
                episodes = []
                lines = fh.readlines()
                header = json.loads(lines[0])
                headers.append(header)
                for line in lines[1:]:
                    episode = json.loads(line)
                    episodes.append(episode)
                df = pandas.DataFrame(episodes)
            else:
                assert 0, 'unreachable'
            df['t'] += header['t_start']
        dfs.append(df)
    df = pandas.concat(dfs)
    df.sort_values('t', inplace=True)
    df.reset_index(inplace=True)
    df['t'] -= min(header['t_start'] for header in headers)
    df.headers = headers # HACK to preserve backwards compatibility
    return df

def test_monitor():
    env = gym.make("CartPole-v1")
    env.seed(0)
    mon_file = "/tmp/baselines-test-%s.monitor.csv" % uuid.uuid4()
    menv = Monitor(env, mon_file)
    menv.reset()
    for _ in range(1000):
        _, _, done, _ = menv.step(0)
        if done:
            menv.reset()

    f = open(mon_file, 'rt')

    firstline = f.readline()
    assert firstline.startswith('#')
    metadata = json.loads(firstline[1:])
    assert metadata['env_id'] == "CartPole-v1"
    assert set(metadata.keys()) == {'env_id', 'gym_version', 't_start'},  "Incorrect keys in monitor metadata"

    last_logline = pandas.read_csv(f, index_col=None)
    assert set(last_logline.keys()) == {'l', 't', 'r'}, "Incorrect keys in monitor logline"
    f.close()
    os.remove(mon_file)



import numpy as np

import torchvision.transforms.functional as F

import vel.api.data as data


class ToTensor(data.Augmentation):
    def __init__(self, mode='x', tags=None):
        super().__init__(mode, tags)

    def __call__(self, datum):
        return F.to_tensor(datum)

    def denormalize(self, datum):
        return np.transpose(datum.numpy(), (1, 2, 0))


def create(mode='x', tags=None):
    return ToTensor(mode, tags)



class Schedule:
    """ A schedule class encoding some kind of interpolation of a single value """
    def value(self, progress_indicator):
        """ Value at given progress step """
        raise NotImplementedError



import torch
import numpy as np

from vel.openai.baselines.common.vec_env import VecEnv
from vel.rl.api import Rollout, Trajectories
from vel.rl.api.base import ReplayEnvRollerBase, EnvRollerFactory
from vel.rl.buffers.deque_multi_env_buffer_backend import DequeMultiEnvBufferBackend


class ReplayQEnvRoller(ReplayEnvRollerBase):
    """
    Class calculating env rollouts and storing them in a buffer for experience replay
    Idea behind this class is to store as much as we can as pytorch tensors to minimize tensor copying.
    """

    def __init__(self, environment: VecEnv, device, number_of_steps, discount_factor, buffer_capacity,
                 buffer_initial_size, frame_stack_compensation):
        self._environment = environment
        self.device = device
        self.number_of_steps = number_of_steps
        self.discount_factor = discount_factor
        self.buffer_capacity = buffer_capacity
        self.buffer_initial_size = buffer_initial_size
        self.frame_stack_compensation = frame_stack_compensation

        # Initial observation
        self.last_observation_cpu = self.environment.reset()
        self.last_observation = self._to_tensor(self.last_observation_cpu)

        # Replay buffer
        self.replay_buffer = DequeMultiEnvBufferBackend(
            buffer_capacity=self.buffer_capacity,
            num_envs=self.environment.num_envs,
            observation_space=self.environment.observation_space,
            action_space=self.environment.action_space,
            extra_data={
                'action_logits': np.zeros(
                    (self.buffer_capacity, self.environment.num_envs, self.environment.action_space.n), dtype=np.float32
                )
            },
            frame_stack_compensation=self.frame_stack_compensation is not None
        )

    @property
    def environment(self):
        """ Return environment of this env roller """
        return self._environment

    def _to_tensor(self, numpy_array):
        """ Convert numpy array to a tensor """
        return torch.from_numpy(numpy_array).to(self.device)

    @torch.no_grad()
    def rollout(self, batch_info, model) -> Rollout:
        """ Calculate env rollout """
        observation_accumulator = []  # Device tensors
        action_accumulator = []  # Device tensors
        logprob_accumulator = []  # Device tensors
        done_accumulator = []  # Device tensors
        reward_accumulator = []  # Device tensors
        episode_information = []  # Python objects

        for step_idx in range(self.number_of_steps):
            step = model.step(self.last_observation)
            actions = step['actions']

            observation_accumulator.append(self.last_observation)
            action_accumulator.append(actions)

            logprobs = step['logprobs']
            logprob_accumulator.append(logprobs)

            actions_numpy = actions.detach().cpu().numpy()
            new_obs, new_rewards, new_dones, new_infos = self.environment.step(actions_numpy)

            # Store rollout in the experience replay buffer
            self.replay_buffer.store_transition(
                frame=self.last_observation_cpu,
                action=actions_numpy,
                reward=new_rewards,
                done=new_dones,
                extra_info={
                    'action_logits': logprobs.detach().cpu().numpy(),
                }
            )

            # Done is flagged true when the episode has ended AND the frame we see is already a first frame from the
            # Next episode
            self.last_observation_cpu = new_obs[:]
            self.last_observation = self._to_tensor(self.last_observation_cpu)

            done_accumulator.append(self._to_tensor(new_dones.astype(np.float32)))
            reward_accumulator.append(self._to_tensor(new_rewards.astype(np.float32)))

            episode_information.append(new_infos)

        final_values = model.value(self.last_observation)

        observations_buffer = torch.stack(observation_accumulator)
        rewards_buffer = torch.stack(reward_accumulator)
        actions_buffer = torch.stack(action_accumulator)
        dones_buffer = torch.stack(done_accumulator)
        action_logit_buffer = torch.stack(logprob_accumulator)

        return Trajectories(
            num_steps=self.number_of_steps,
            num_envs=self.environment.num_envs,
            environment_information=episode_information,
            transition_tensors={
                'observations': observations_buffer,
                'rewards': rewards_buffer,
                'dones': dones_buffer,
                'actions': actions_buffer,
                'logprobs': action_logit_buffer,
            },
            rollout_tensors={
                'final_estimated_values': final_values
            }
        )

    def is_ready_for_sampling(self) -> bool:
        """ If buffer is ready for drawing samples from it (usually checks if there is enough data) """
        return self.replay_buffer.current_size >= self.buffer_initial_size

    @torch.no_grad()
    def sample(self, batch_info, model):
        """ Sample experience from replay buffer and return a batch """
        rollout_idx = self.replay_buffer.sample_batch_rollout(
            rollout_length=self.number_of_steps, history_length=self.frame_stack_compensation
        )

        rollout = self.replay_buffer.get_rollout(
            rollout_idx, rollout_length=self.number_of_steps, history_length=self.frame_stack_compensation
        )

        return Trajectories(
            num_steps=self.number_of_steps,
            num_envs=self.environment.num_envs,
            environment_information=None,
            transition_tensors={
                'observations': self._to_tensor(rollout['states']),
                'dones': self._to_tensor(rollout['dones'].astype(np.float32)),
                'rewards': self._to_tensor(rollout['rewards']),
                'actions': self._to_tensor(rollout['actions']),
                'logprobs': self._to_tensor(rollout['action_logits'])
            },
            rollout_tensors={
                'final_estimated_values': model.value(self._to_tensor(rollout['states+1'][-1]))
            }
        )


class ReplayQEnvRollerFactory(EnvRollerFactory):
    """ Factory for the StepEnvRoller """
    def __init__(self, buffer_capacity, buffer_initial_size, number_of_steps, frame_stack_compensation=None):
        self.buffer_capacity = buffer_capacity
        self.buffer_initial_size = buffer_initial_size
        self.frame_stack_compensation = frame_stack_compensation
        self.number_of_steps = number_of_steps

    def instantiate(self, environment, device, settings):
        return ReplayQEnvRoller(
            environment, device, self.number_of_steps, settings.discount_factor,
            self.buffer_capacity, self.buffer_initial_size,
            frame_stack_compensation=self.frame_stack_compensation
        )


def create(buffer_capacity, buffer_initial_size, number_of_steps, frame_stack_compensation=None):
    return ReplayQEnvRollerFactory(
        buffer_capacity=buffer_capacity,
        number_of_steps=number_of_steps,
        buffer_initial_size=buffer_initial_size,
        frame_stack_compensation=frame_stack_compensation
    )



"""
Code based loosely on implementation:
https://github.com/openai/baselines/blob/master/baselines/ppo2/policies.py

Under MIT license.
"""
import numpy as np

import torch
import torch.nn as nn
import torch.nn.init as init
import torch.nn.functional as F

import vel.util.network as net_util

from vel.api.base import LinearBackboneModel, ModelFactory


class NatureCnnSmall(LinearBackboneModel):
    """ Neural network as defined in the paper 'Human-level control through deep reinforcement learning'"""
    def __init__(self, input_width, input_height, input_channels, output_dim=128):
        super().__init__()

        self._output_dim = output_dim

        self.conv1 = nn.Conv2d(
            in_channels=input_channels,
            out_channels=8,
            kernel_size=(8, 8),
            stride=4
        )

        self.conv2 = nn.Conv2d(
            in_channels=8,
            out_channels=16,
            kernel_size=(4, 4),
            stride=2
        )

        self.final_width = net_util.convolutional_layer_series(input_width, [
            (8, 0, 4),
            (4, 0, 2),
        ])

        self.final_height = net_util.convolutional_layer_series(input_height, [
            (8, 0, 4),
            (4, 0, 2),
        ])

        self.linear_layer = nn.Linear(
            self.final_width * self.final_height * 16,
            self.output_dim
        )

    @property
    def output_dim(self) -> int:
        """ Final dimension of model output """
        return self._output_dim

    def reset_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                # init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                init.orthogonal_(m.weight, gain=np.sqrt(2))
                init.constant_(m.bias, 0.0)
            elif isinstance(m, nn.Linear):
                # init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                init.orthogonal_(m.weight, gain=np.sqrt(2))
                init.constant_(m.bias, 0.0)

    def forward(self, image):
        result = image.permute(0, 3, 1, 2).contiguous().type(torch.float) / 255.0
        result = F.relu(self.conv1(result))
        result = F.relu(self.conv2(result))
        flattened = result.view(result.size(0), -1)
        return F.relu(self.linear_layer(flattened))


def create(input_width, input_height, input_channels=1):
    def instantiate(**_):
        return NatureCnnSmall(input_width=input_width, input_height=input_height, input_channels=input_channels)

    return ModelFactory.generic(instantiate)


NatureCnnSmallFactory = create



import cv2
import numpy as np
import os.path
import pathlib
import sys
import torch
import tqdm
import typing

from vel.api import ModelConfig, TrainingInfo
from vel.api.base import Storage, ModelFactory
from vel.rl.api.base import EnvFactory
from vel.openai.baselines.common.atari_wrappers import FrameStack


class RecordMovieCommand:
    """ Record environment playthrough as a game  """
    def __init__(self, model_config: ModelConfig, env_factory: EnvFactory, model_factory: ModelFactory,
                 storage: Storage, videoname: str, takes: int, frame_history: typing.Optional[int],
                 fps: int, sample_args: typing.Optional[dict] = None):
        self.model_config = model_config
        self.model_factory = model_factory
        self.env_factory = env_factory
        self.storage = storage
        self.takes = takes
        self.videoname = videoname
        self.frame_history = frame_history
        self.sample_args = sample_args if sample_args is not None else {}
        self.fps = fps

    def run(self):
        device = torch.device(self.model_config.device)

        env = self.env_factory.instantiate(preset='raw')

        if self.frame_history:
            env = FrameStack(env, self.frame_history)

        model = self.model_factory.instantiate(action_space=env.action_space).to(device)

        training_info = TrainingInfo(start_epoch_idx=self.storage.last_epoch_idx(), run_name=self.model_config.run_name)
        self.storage.resume(training_info, model)

        model.eval()

        for i in range(self.takes):
            self.record_take(model, env, device, takenumber=i+1)

    @torch.no_grad()
    def record_take(self, model, env_instance, device, takenumber):
        frames = []

        observation = env_instance.reset()

        frames.append(env_instance.render('rgb_array'))

        print("Evaluating environment...")

        while True:
            observation_array = np.expand_dims(np.array(observation), axis=0)
            observation_tensor = torch.from_numpy(observation_array).to(device)
            actions = model.step(observation_tensor, **self.sample_args)['actions']
            actions = actions.detach().cpu().numpy()

            observation, reward, done, epinfo = env_instance.step(actions[0])

            frames.append(env_instance.render('rgb_array'))

            if 'episode' in epinfo:
                # End of an episode
                break

        takename = self.model_config.output_dir('videos', self.model_config.run_name, self.videoname.format(takenumber))
        pathlib.Path(os.path.dirname(takename)).mkdir(parents=True, exist_ok=True)

        fourcc = cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')
        video = cv2.VideoWriter(takename, fourcc, self.fps, (frames[0].shape[1], frames[0].shape[0]))

        for i in tqdm.trange(len(frames), file=sys.stdout):
            video.write(cv2.cvtColor(frames[i], cv2.COLOR_RGB2BGR))

        video.release()
        print(f"Written {takename}")


def create(model_config, model, env, storage, takes, videoname, frame_history=None, fps=30, sample_args=None):
    return RecordMovieCommand(
        model_config=model_config,
        model_factory=model,
        env_factory=env,
        storage=storage,
        videoname=videoname,
        frame_history=frame_history,
        takes=takes,
        fps=fps,
        sample_args=sample_args
    )



import torch
import torch.nn.functional as F

from vel.api.metrics.averaging_metric import AveragingNamedMetric
from vel.rl.api.base import OptimizerAlgoBase
from vel.math.functions import explained_variance


class A2CPolicyGradient(OptimizerAlgoBase):
    """ Simplest policy gradient - calculate loss as an advantage of an actor versus value function """
    def __init__(self, entropy_coefficient, value_coefficient, max_grad_norm):
        super().__init__(max_grad_norm)

        self.entropy_coefficient = entropy_coefficient
        self.value_coefficient = value_coefficient

    def calculate_gradient(self, batch_info, device, model, rollout):
        """ Calculate loss of the supplied rollout """
        evaluator = model.evaluate(rollout)

        # Use evaluator interface to get the what we are interested in from the model
        advantages = evaluator.get('rollout:estimated_advantages')
        returns = evaluator.get('rollout:estimated_returns')
        rollout_values = evaluator.get('rollout:estimated_values')

        logprobs = evaluator.get('model:action:logprobs')
        values = evaluator.get('model:estimated_values')
        entropy = evaluator.get('model:entropy')

        # Actual calculations. Pretty trivial
        policy_loss = -torch.mean(advantages * logprobs)
        value_loss = 0.5 * F.mse_loss(values, returns)
        policy_entropy = torch.mean(entropy)

        loss_value = (
            policy_loss - self.entropy_coefficient * policy_entropy + self.value_coefficient * value_loss
        )

        loss_value.backward()

        return {
            'policy_loss': policy_loss.item(),
            'value_loss': value_loss.item(),
            'policy_entropy': policy_entropy.item(),
            'advantage_norm': torch.norm(advantages).item(),
            'explained_variance': explained_variance(returns, rollout_values)
        }

    def metrics(self) -> list:
        """ List of metrics to track for this learning process """
        return [
            AveragingNamedMetric("value_loss"),
            AveragingNamedMetric("policy_entropy"),
            AveragingNamedMetric("policy_loss"),
            AveragingNamedMetric("grad_norm"),
            AveragingNamedMetric("advantage_norm"),
            AveragingNamedMetric("explained_variance")
        ]


def create(entropy_coefficient, value_coefficient, max_grad_norm):
    return A2CPolicyGradient(entropy_coefficient, value_coefficient, max_grad_norm)



"""
Code based on:
https://github.com/fastai/fastai/blob/master/fastai/transforms.py
"""

import vel.api.data as data


class CenterCrop(data.Augmentation):
    """ A class that represents a Center Crop.

    This transforms (optionally) transforms x,y at with the same parameters.
    Arguments
    ---------
        sz: int
            size of the crop.
        tfm_y : TfmType
            type of y transformation.
    """
    def __init__(self, size, mode='x', tags=None):
        super().__init__(mode, tags)

        self.size = size

    def __call__(self, x):
        return data.center_crop(x, self.size)


def create(size, mode='x', tags=None):
    return CenterCrop(size, mode, tags)



import itertools as it

import gym

from vel.api.base import LinearBackboneModel, Model, ModelFactory
from vel.rl.api import Rollout, Evaluator
from vel.rl.modules.action_head import ActionHead
from vel.rl.modules.value_head import ValueHead
from vel.rl.models.policy_gradient_model import PolicyGradientEvaluator


class PolicyGradientModelSeparate(Model):
    """ For a policy gradient algorithm we need set of custom heads for our model """

    def __init__(self, policy_backbone: LinearBackboneModel, value_backbone: LinearBackboneModel,
                 action_space: gym.Space):
        super().__init__()

        self.policy_backbone = policy_backbone
        self.value_backbone = value_backbone

        self.action_head = ActionHead(
            action_space=action_space,
            input_dim=self.policy_backbone.output_dim
        )

        self.value_head = ValueHead(input_dim=self.value_backbone.output_dim)

    def reset_weights(self):
        """ Initialize properly model weights """
        self.policy_backbone.reset_weights()
        self.value_backbone.reset_weights()

        self.action_head.reset_weights()
        self.value_head.reset_weights()

    def forward(self, observations):
        """ Calculate model outputs """
        policy_base_output = self.policy_backbone(observations)
        value_base_output = self.value_backbone(observations)

        action_output = self.action_head(policy_base_output)
        value_output = self.value_head(value_base_output)

        return action_output, value_output

    def step(self, observation, argmax_sampling=False):
        """ Select actions based on model's output """
        policy_params, values = self(observation)
        actions = self.action_head.sample(policy_params, argmax_sampling=argmax_sampling)

        # log likelihood of selected action
        logprobs = self.action_head.logprob(actions, policy_params)

        return {
            'actions': actions,
            'values': values,
            'logprobs': logprobs
        }

    def policy_parameters(self):
        """ Parameters of policy """
        return it.chain(self.policy_backbone.parameters(), self.action_head.parameters())

    def logprob(self, action_sample, policy_params):
        """ Calculate - log(prob) of selected actions """
        return self.action_head.logprob(action_sample, policy_params)

    def value(self, observation):
        """ Calculate only value head for given state """
        base_output = self.value_backbone(observation)
        value_output = self.value_head(base_output)
        return value_output

    def policy(self, observation):
        """ Calculate only action head for given state """
        policy_base_output = self.policy_backbone(observation)
        policy_params = self.action_head(policy_base_output)
        return policy_params

    def evaluate(self, rollout: Rollout) -> Evaluator:
        """ Evaluate model on a rollout """
        return PolicyGradientEvaluator(self, rollout)

    def entropy(self, policy_params):
        """ Entropy of a probability distribution """
        return self.action_head.entropy(policy_params)

    def kl_divergence(self, pd_q, pd_p):
        """ Calculate KL-divergence between two probability distributions """
        return self.action_head.kl_divergence(pd_q, pd_p)


class PolicyGradientModelSeparateFactory(ModelFactory):
    """ Factory  class for policy gradient models """
    def __init__(self, policy_backbone: ModelFactory, value_backbone: ModelFactory):
        self.policy_backbone = policy_backbone
        self.value_backbone = value_backbone

    def instantiate(self, **extra_args):
        """ Instantiate the model """
        policy_backbone = self.policy_backbone.instantiate(**extra_args)
        value_backbone = self.value_backbone.instantiate(**extra_args)

        return PolicyGradientModelSeparate(policy_backbone, value_backbone, extra_args['action_space'])


def create(policy_backbone: ModelFactory, value_backbone: ModelFactory):
    return PolicyGradientModelSeparateFactory(
        policy_backbone=policy_backbone,
        value_backbone=value_backbone
    )



# import torch
# import torch.utils.data as tdata
#
# import vel.api.data as wdata
#
#
# class TrainTTALoader:
#     def __init__(self, n_augmentations, batch_size, data_source, augmentations, num_workers):
#         self.n_augmentations = n_augmentations
#         self.data_source = data_source
#         self.augmentations = augmentations
#
#         self.val_ds = wdata.DataFlow(self.data_source, augmentations, tag='val')
#         self.train_ds = wdata.DataFlow(self.data_source, augmentations, tag='train')
#
#         self.val_loader = tdata.DataLoader(
#             self.val_ds, batch_size=batch_size, shuffle=False, num_workers=num_workers
#         )
#
#         self.train_loader = tdata.DataLoader(
#             self.train_ds, batch_size=batch_size, shuffle=False, num_workers=num_workers
#         )
#
#     def __len__(self):
#         return (1 + self.n_augmentations) * len(self.val_loader)
#
#     def __iter__(self):
#         iterlist = [iter(self.val_loader)]
#
#         for _ in range(self.n_augmentations):
#             iterlist.append(iter(self.train_loader))
#
#         for _ in range(len(self.val_loader)):
#             for iterator in iterlist:
#                 yield next(iterator)
#
#
# class TrainTTAAccumulator:
#     def __init__(self, metric_accumulator, n_augmentations, data_source):
#         self.metric_accumulator = metric_accumulator
#
#         self.source_elements = len(data_source)
#         self.n_augmentations = n_augmentations
#
#         self.data = None
#         self.target = None
#
#         self.accumulated_output = []
#         self.accumulated_context = []
#
#         self.index = 0
#
#     # def calculate(self, data, target, output, context):
#     def calculate(self, data_dict):
#         """ Accumulate results """
#         data = data_dict['data']
#         target = data_dict['target']
#         output = data_dict['output']
#
#         if self.index == 0:
#             self.data = data
#
#         self.target = target
#
#         self.accumulated_output.append(output)
#         self.accumulated_context.append(context)
#
#         self.index += 1
#
#         if self.index == (1 + self.n_augmentations):
#             new_output = torch.mean(torch.stack(self.accumulated_output, dim=-1), dim=-1)
#             new_context = {
#                 k: torch.mean(torch.stack([c[k] for c in self.accumulated_context], dim=-1), dim=-1) for k in context.keys()
#             }
#
#             self.metric_accumulator.calculate(self.data, self.target, new_output, new_context)
#
#             self.index = 0
#             self.data = None
#             self.target = None
#             self.accumulated_output = []
#             self.accumulated_context = []
#
#
# class TrainTTA:
#     """ Test time augmentation that generates additional samples according to the training set augmentations """
#     def __init__(self, n_augmentations):
#         self.n_augmentations = n_augmentations
#
#     def loader(self, data_source, augmentations, batch_size, num_workers):
#         """ Return loader for the test-time-augmentation set """
#         return TrainTTALoader(
#             n_augmentations=self.n_augmentations,
#             batch_size=batch_size,
#             data_source=data_source,
#             augmentations=augmentations,
#             num_workers=num_workers
#         )
#
#     def accumulator(self, metric_accumulator, val_source):
#         """ Reset internal state """
#         return TrainTTAAccumulator(metric_accumulator, self.n_augmentations, val_source)
#
#
# def create(n_augmentations):
#     return TrainTTA(n_augmentations)



"""
Code based loosely on the implementation:
https://github.com/openai/baselines/blob/master/baselines/common/models.py

Under MIT license.
"""
import typing
import numpy as np

import torch.nn as nn
import torch.nn.init as init

import vel.util.network as net_util

from vel.api.base import LinearBackboneModel, ModelFactory


class MLP(LinearBackboneModel):
    """ Simple Multi-Layer-Perceptron network """
    def __init__(self, input_length: int, hidden_layers: typing.List[int], activation: str='tanh',
                 normalization: typing.Optional[str]=None):
        super().__init__()

        self.input_length = input_length
        self.hidden_layers = hidden_layers
        self.activation = activation
        self.normalization = normalization

        layer_objects = []
        layer_sizes = zip([input_length] + hidden_layers, hidden_layers)

        for input_size, output_size in layer_sizes:
            layer_objects.append(nn.Linear(input_size, output_size))

            if self.normalization:
                layer_objects.append(net_util.normalization(normalization)(output_size))

            layer_objects.append(net_util.activation(activation)())

        self.model = nn.Sequential(*layer_objects)
        self.hidden_units = hidden_layers[-1] if hidden_layers else input_length

    @property
    def output_dim(self) -> int:
        """ Final dimension of model output """
        return self.hidden_units

    def reset_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                # init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                init.orthogonal_(m.weight, gain=np.sqrt(2))
                init.constant_(m.bias, 0.0)

    def forward(self, input_data):
        input_data = input_data.float()
        return self.model(input_data)


def create(input_length, hidden_layers, activation='tanh', normalization=None):
    def instantiate(**_):
        return MLP(
            input_length=input_length,
            hidden_layers=hidden_layers,
            activation=activation,
            normalization=normalization
        )

    return ModelFactory.generic(instantiate)


MLPFactory = create
